from pathlib import Path

import httpx
from kurra.db import upload, clear_graph
from kurra.sparql import query
from kurra.utils import load_graph
from rdflib import Graph
from rdflib import URIRef
from rdflib.namespace import SDO

from prezmanifest.definednamespaces import MRR
from prezmanifest.utils import (
    VersionIndicatorComparison,
    store_remote_artifact_locally,
    update_local_artifact,
    absolutise_path,
)
from prezmanifest.utils import (
    get_manifest_paths_and_graph,
    denormalise_artifacts,
    which_is_more_recent,
)


def sync(
    manifest: Path | tuple[Path, Path, Graph],
    sparql_endpoint: str = None,
    http_client: httpx.Client = httpx.Client(),
    update_remote: bool = True,
    update_local: bool = True,
    add_remote: bool = True,
    add_local: bool = True,
) -> dict:
    manifest_path, manifest_root, manifest_graph = get_manifest_paths_and_graph(
        manifest
    )

    sync_status = {}
    # For each Artifact in the Manifest
    artifacts = denormalise_artifacts((manifest_path, manifest_root, manifest_graph))
    local_entities = [v["main_entity"] for k, v in artifacts.items()]

    cat_iri = None
    cat_artifact_path = None
    for k, v in artifacts.items():
        if v["role"] in [MRR.ResourceData, MRR.CatalogueData]:
            # save cat_iri for later
            if v["role"] in MRR.CatalogueData:
                cat_iri = v["main_entity"]
                cat_artifact_path = absolutise_path(k, manifest_root)

            # See if each is known remotely (via Main Entity Graph IRI)
            known = query(
                sparql_endpoint,
                "ASK {GRAPH <xxx> {?s ?p ?o}}".replace("xxx", str(v["main_entity"])),
                http_client,
                return_python=True,
                return_bindings_only=True,
            )
            # If not known by graph IRI, just check if it's the catalogue (+ "-catalogue" to IRI)
            if not known and v["role"] == MRR.CatalogueData:
                known = query(
                    sparql_endpoint,
                    "ASK {GRAPH <xxx> {?s ?p ?o}}".replace(
                        "xxx", str(v["main_entity"] + "-catalogue")
                    ),
                    http_client,
                    return_python=True,
                    return_bindings_only=True,
                )

            # If known, compare it
            if known:
                replace = which_is_more_recent(
                    v,
                    sparql_endpoint,
                    http_client,
                )
                if replace == VersionIndicatorComparison.First:
                    direction = "upload"
                elif replace == VersionIndicatorComparison.Second:
                    direction = "download"
                elif replace == VersionIndicatorComparison.Neither:
                    direction = "same"
                elif VersionIndicatorComparison.CantCalculate:
                    direction = "upload"
            else:  # not known at remote location so forward sync - upload
                direction = "add-remotely"

            sync_status[str(k)] = {
                "main_entity": v["main_entity"],
                "direction": direction,
            }

    # Check for things at remote not known in local
    q = """
        PREFIX dcterms: <http://purl.org/dc/terms/>
        PREFIX schema: <https://schema.org/>
    
        SELECT ?p
        WHERE {
            GRAPH ?g {
                <xxx> schema:hasPart|dcterms:hasPart ?p
            }
        }
        """.replace(
        "xxx", str(cat_iri)
    )
    for x in query(
        sparql_endpoint, q, http_client, return_python=True, return_bindings_only=True
    ):
        remote_entity = URIRef(x["p"]["value"])
        if remote_entity not in local_entities:
            sync_status[str(remote_entity)] = {
                "main_entity": URIRef(remote_entity),
                "direction": "add-locally",
            }

    update_remote_catalogue = False
    for k, v in sync_status.items():
        if update_remote and v["direction"] == "upload":
            clear_graph(sparql_endpoint, v["main_entity"], http_client)
            upload(sparql_endpoint, Path(k), v["main_entity"], False, http_client)

        if add_remote and v["direction"] == "add-remotely":
            # no need to clear_graph() as this asset doesn't exist remotely
            upload(sparql_endpoint, Path(k), v["main_entity"], False, http_client)
            update_remote_catalogue = True

        if add_local and v["direction"] == "add-locally":
            updated_local_manifest = store_remote_artifact_locally(
                (manifest_path, manifest_root, manifest_graph),
                sparql_endpoint,
                v["main_entity"],
                http_client,
            )

            updated_local_manifest.bind("mrr", "https://prez.dev/ManifestResourceRoles")
            updated_local_manifest.serialize(
                destination=manifest_path, format="longturtle"
            )
            cat = load_graph(cat_artifact_path)
            cat.add((cat_iri, SDO.hasPart, URIRef(v["main_entity"])))
            cat.serialize(destination=cat_artifact_path, format="longturtle")

        if update_local and v["direction"] == "download":
            update_local_artifact(
                (manifest_path, manifest_root, manifest_graph),
                Path(k),
                sparql_endpoint,
                v["main_entity"],
                http_client,
            )

    if update_remote_catalogue:
        clear_graph(sparql_endpoint, cat_iri, http_client)
        upload(
            sparql_endpoint,
            cat_artifact_path,
            cat_iri,
            False,
            http_client,
        )

    return sync_status



from pathlib import Path

from typer.testing import CliRunner

from prezmanifest import validate
from prezmanifest.cli import app
from prezmanifest.validator import ManifestValidationError

runner = CliRunner()


def test_validator_valid():
    assert validate(Path(__file__).parent / "demo-vocabs" / "manifest.ttl")


def test_validator_invalid_01():
    try:
        validate(Path(__file__).parent / "demo-vocabs" / "manifest-invalid-01.ttl")
    except ManifestValidationError as e:
        assert "Manifest Shapes invalid:" in str(e)


def test_validator_invalid_03():
    try:
        validate(Path(__file__).parent / "demo-vocabs" / "manifest-invalid-02.ttl")
    except ManifestValidationError as e:
        assert str(e) == "The content link vocabz/*.ttl is not a directory"


def test_validator_invalid_02():
    try:
        validate(Path(__file__).parent / "demo-vocabs" / "manifest-invalid-03.ttl")
    except ManifestValidationError as e:
        assert (
            str(e)
            == "Remote content link non-resolving: https://raw.githubusercontent.com/RDFLib/prez/refs/heads/main/prez/reference_data/profiles/ogc_records_profile.ttlx"
        )


def test_validator_valid_multi():
    assert validate(Path(__file__).parent / "demo-vocabs" / "manifest-multi.ttl")


def test_validator_valid_main_entity():
    assert validate(Path(__file__).parent / "demo-vocabs" / "manifest-mainEntity.ttl")


def test_validator_invalid_main_entity():
    try:
        validate(
            Path(__file__).parent / "demo-vocabs" / "manifest-mainEntity-invalid.ttl"
        )
    except ManifestValidationError as e:
        assert "N04" in str(e)


def test_validator_invalid_main_entity2():
    try:
        validate(
            Path(__file__).parent / "demo-vocabs" / "manifest-mainEntity-invalid2.ttl"
        )
    except ManifestValidationError as e:
        assert "N04" in str(e)


def test_validator_valid_conformance():
    assert validate(Path(__file__).parent / "demo-vocabs" / "manifest-conformance.ttl")


def test_validator_valid_conformance_local():
    assert validate(
        Path(__file__).parent / "demo-vocabs" / "manifest-conformance-local.ttl"
    )


def test_validator_invalid_conformance_local():
    try:
        validate(
            Path(__file__).parent
            / "demo-vocabs"
            / "manifest-conformance-local-invalid.ttl"
        )
    except ManifestValidationError as e:
        assert "Message: Requirement 2.1.4, 2.2.1 or 2.3.1" in str(e)


def test_validator_valid_conformance_all():
    try:
        validate(Path(__file__).parent / "demo-vocabs" / "manifest-conformance-all.ttl")
    except ManifestValidationError as e:
        assert "Results (6)" in str(e)

    # language-test.ttl is known to have 6 errors according to VocPub 4.10, image-test.ttl none


def test_validator_invalid_conformance_all():
    try:
        validate(
            Path(__file__).parent
            / "demo-vocabs"
            / "manifest-conformance-all-local-invalid.ttl"
        )
    except ManifestValidationError as e:
        assert "Results (1)" in str(e)


def test_own_validator():
    from rdflib import Namespace

    GN = Namespace("https://linked.data.gov.au/def/gn/")

    m = Path(__file__).parent / "validator/manifest-conformance-own.ttl"
    try:
        vg = validate(m)
    except ManifestValidationError as e:
        assert "Results (5)" in str(e)


def test_validator_cli():
    try:
        runner.invoke(
            app,
            [
                "validate",
                Path(__file__).parent / "demo-vocabs" / "manifest-invalid-01.ttl",
            ],
        )
    except ManifestValidationError as e:
        assert "MinCountConstraintComponent" in str(e)



"""
Either creates an n-quads files containing the content of a Manifest file or uploads the content to Fuseki.

It creates:

 1. A Named Graph for each resource using the item's IRI as the graph IRI
 2. A Named Graph for the catalogue, either using the catalogue's IRI as the graph IRI + "-catalogue" if given, or by making one up - a Blank Node
 3. All the triples in resources with roles mrr:CompleteCatalogueAndResourceLabels & mrr:IncompleteCatalogueAndResourceLabels within a Named Graph with IRI <https://background>
 4. An Olis Virtual Graph, <https://olis.dev/VirtualGraph> object using the catalogue IRI, if give, which is as an alias for all the Named Graphs from 1., 2. & 3.
 5. Multiple entries in the System Graph - Named Graph with IRI <https://olis.dev/SystemGraph> - for each Named and the Virtual Graph from 1., 2. & 3.

Run this script with the -h flag for more help, i.e. ~$ python loader.py -h
"""

import logging
import sys
from enum import Enum
from getpass import getpass
from pathlib import Path

import httpx
from kurra.db import upload
from kurra.file import export_quads, make_dataset
from kurra.utils import load_graph
from rdflib import DCTERMS, PROF, RDF, SDO, SKOS, Dataset, Graph, URIRef

from prezmanifest.definednamespaces import MRR, OLIS
from prezmanifest.utils import (
    KNOWN_ENTITY_CLASSES,
    get_files_from_artifact,
    get_catalogue_iri_from_manifest,
)
from prezmanifest.utils import get_manifest_paths_and_graph


class ReturnDatatype(str, Enum):
    graph = "graph"
    dataset = "dataset"
    none = None


def load(
    manifest: Path | tuple[Path, Path, Graph],
    sparql_endpoint: str = None,
    sparql_username: str = None,
    sparql_password: str = None,
    destination_file: Path = None,
    return_data_type: ReturnDatatype = ReturnDatatype.none,
) -> None | Graph | Dataset:
    """Loads a catalogue of data from a prezmanifest file, whose content are valid according to the Prez Manifest Model
    (https://kurrawong.github.io/prez.dev/manifest/) either into a specified quads file in the Trig format, or into a
    given SPARQL Endpoint."""

    # validate and load
    manifest_path, manifest_root, manifest_graph = get_manifest_paths_and_graph(
        manifest
    )

    catalogue_iri_orig = get_catalogue_iri_from_manifest(
        (manifest_path, manifest_root, manifest_graph)
    )
    vg_iri = catalogue_iri_orig
    catalogue_iri = URIRef(str(catalogue_iri_orig) + "-catalogue")

    if not isinstance(return_data_type, ReturnDatatype):
        raise ValueError(
            f"Invalid return_data_type value. Must be one of {', '.join([x for x in ReturnDatatype])}"
        )

    if (
        sparql_endpoint is None
        and destination_file is None
        and return_data_type == ReturnDatatype.none
    ):
        raise ValueError(
            "Either a sparql_endpoint, destination_file or a return_data_type must be specified"
        )

    if return_data_type == ReturnDatatype.dataset:
        dataset_holder = Dataset()

    if return_data_type == ReturnDatatype.graph:
        graph_holder = Graph()

    # establish a reusable client for http requests
    # also allows for basic authentication to be used
    if sparql_endpoint:
        auth = None
        if sparql_username:
            if not sparql_password:
                if not sys.stdin.isatty():
                    # if not possible to prompt for a password
                    raise ValueError(
                        "A password must be given if a sparql username is set"
                    )
                sparql_password = getpass()
            auth = httpx.BasicAuth(sparql_username, sparql_password)
        http_client = httpx.Client(base_url=sparql_endpoint, auth=auth)
    else:
        http_client = None

    def _export(
        data: Graph | Dataset,
        iri,
        http_client: httpx.Client | None,
        sparql_endpoint,
        destination_file,
        return_data_type,
        append=False,
    ):
        if type(data) is Dataset:
            if iri is not None:
                raise ValueError(
                    "If the data is a Dataset, the parameter iri must be None"
                )

            if destination_file is not None:
                export_quads(data, destination_file)
            elif sparql_endpoint is not None:
                for g in data.graphs():
                    if g.identifier != URIRef("urn:x-rdflib:default"):
                        _export(
                            data=g,
                            iri=g.identifier,
                            http_client=http_client,
                            destination_file=None,
                            return_data_type=None,
                        )
            else:
                if return_data_type == "Dataset":
                    return data
                elif return_data_type == "Graph":
                    gx = Graph()
                    for g in data.graphs():
                        if g.identifier != URIRef("urn:x-rdflib:default"):
                            for s, p, o in g.triples((None, None, None)):
                                gx.add((s, p, o))
                    return gx

        elif type(data) is Graph:
            if iri is None:
                raise ValueError(
                    "If the data is a GRaph, the parameter iri must not be None"
                )

            msg = f"exporting {iri} "
            if destination_file is not None:
                msg += f"to file {destination_file} "
                export_quads(make_dataset(data, iri), destination_file)
            elif sparql_endpoint is not None:
                msg += f"to SPARQL Endpoint {sparql_endpoint}"
                upload(
                    sparql_endpoint=sparql_endpoint,
                    file_or_str_or_graph=data,
                    graph_id=iri,
                    append=append,
                    http_client=http_client,
                )
            else:  # returning data
                if return_data_type == ReturnDatatype.dataset:
                    msg += "to Dataset"
                    for s, p, o in data:
                        dataset_holder.add((s, p, o, iri))
                elif return_data_type == ReturnDatatype.graph:
                    msg += "to Graph"
                    for s, p, o in data:
                        graph_holder.add((s, p, o))

            logging.info(msg)

    count = 0
    if sparql_endpoint is not None:
        count += 1

    if destination_file is not None:
        count += 1

    if return_data_type != ReturnDatatype.none:
        count += 1

    if count != 1:
        raise ValueError(
            "You must specify exactly 1 of sparql_endpoint, destination_file or return_data_type",
        )

    vg = Graph()

    for s, o in manifest_graph.subject_objects(PROF.hasResource):
        for role in manifest_graph.objects(o, PROF.hasRole):
            # The catalogue - must be processed first
            if role == MRR.CatalogueData:
                for artifact in manifest_graph.objects(o, PROF.hasArtifact):
                    # load the Catalogue, determine the Virtual Graph & Catalogue IRIs
                    # and fail if we can't see a Catalogue object
                    catalogue_graph = load_graph(manifest_root / artifact)

                    if vg_iri is None:
                        raise ValueError(
                            "ERROR: Could not create a Virtual Graph as no Catalog found in the Catalogue data"
                        )

                    # add to the System Graph
                    vg.add((vg_iri, RDF.type, OLIS.VirtualGraph))
                    vg.add((vg_iri, OLIS.isAliasFor, catalogue_iri))
                    vg_name = catalogue_graph.value(  # type: ignore
                        subject=vg_iri,
                        predicate=SDO.name | DCTERMS.title | SKOS.prefLabel,
                    ) or str(vg_iri)
                    vg.add((vg_iri, SDO.name, vg_name))

                    # export the Catalogue data
                    _export(
                        data=catalogue_graph,
                        iri=catalogue_iri,
                        http_client=http_client,
                        sparql_endpoint=sparql_endpoint,
                        destination_file=destination_file,
                        return_data_type=return_data_type,
                    )

    # non-catalogue resources
    for s, o in manifest_graph.subject_objects(PROF.hasResource):
        for role in manifest_graph.objects(o, PROF.hasRole):
            # The data files & background - must be processed after Catalogue
            if role in [
                MRR.CompleteCatalogueAndResourceLabels,
                MRR.IncompleteCatalogueAndResourceLabels,
                MRR.ResourceData,
            ]:
                for artifact in manifest_graph.objects(o, PROF.hasArtifact):
                    for f in get_files_from_artifact(
                        (manifest_path, manifest_root, manifest_graph), artifact
                    ):
                        if str(f.name).endswith(".ttl"):
                            try:
                                fg = Graph().parse(f)
                            except Exception as e:
                                raise ValueError(
                                    f"Could not load file {f}. Error is {e}"
                                )

                            # fg.bind("rdf", RDF)

                            if role == MRR.ResourceData:
                                resource_iri = fg.value(
                                    subject=artifact, predicate=SDO.mainEntity
                                )
                                if resource_iri is None:
                                    for entity_class in KNOWN_ENTITY_CLASSES:
                                        v = fg.value(
                                            predicate=RDF.type, object=entity_class
                                        )
                                        if v is not None:
                                            resource_iri = v

                            if role in [
                                MRR.CompleteCatalogueAndResourceLabels,
                                MRR.IncompleteCatalogueAndResourceLabels,
                            ]:
                                resource_iri = URIRef("http://background")

                            if resource_iri is None:
                                raise ValueError(
                                    f"Could not determine Resource IRI for file {f}"
                                )

                            vg.add((vg_iri, OLIS.isAliasFor, resource_iri))

                            # export one Resource
                            _export(
                                data=fg,
                                iri=resource_iri,
                                http_client=http_client,
                                sparql_endpoint=sparql_endpoint,
                                destination_file=destination_file,
                                return_data_type=return_data_type,
                            )
                        elif str(f.name).endswith(".trig"):
                            d = Dataset()
                            d.parse(f)
                            for g in d.graphs():
                                if g.identifier != URIRef("urn:x-rdflib:default"):
                                    vg.add((vg_iri, OLIS.isAliasFor, g.identifier))
                            _export(
                                data=d,
                                iri=None,
                                http_client=http_client,
                                sparql_endpoint=sparql_endpoint,
                                destination_file=destination_file,
                                return_data_type=return_data_type,
                            )

        # export the System Graph
        _export(
            data=vg,
            iri=OLIS.SystemGraph,
            http_client=http_client,
            sparql_endpoint=sparql_endpoint,
            destination_file=destination_file,
            return_data_type=return_data_type,
            append=True,
        )

    if return_data_type == ReturnDatatype.dataset:
        return dataset_holder
    elif return_data_type == ReturnDatatype.graph:
        return graph_holder
    else:  # return_data_type is None:
        pass  # return nothing



from pathlib import Path
from typing import Annotated

import typer

from prezmanifest.labeller import LabellerOutputTypes, label
from prezmanifest.utils import make_httpx_client

app = typer.Typer(
    help="Discover labels missing from data in a in a Prez Manifest and patch them"
)


@app.command(
    name="iris",
    help="Find all the IRIs of objects in the Manifest's resources without labels",
)
def iris_command(
    manifest: Path = typer.Argument(
        ..., help="The path of the Prez Manifest file to be labelled"
    ),
    context: str = typer.Argument(
        None,
        help="The path of an RDF file, a directory of RDF files or the URL of a SPARQL endpoint from which to obtain labels",
    ),
    username: Annotated[
        str, typer.Option("--username", "-u", help="SPARQL Endpoint username")
    ] = None,
    password: Annotated[
        str, typer.Option("--password", "-p", help="SPARQL Endpoint password")
    ] = None,
) -> None:
    for iri in label(
        manifest,
        LabellerOutputTypes.iris,
        context,
        make_httpx_client(username, password),
    ):
        print(str(iri))


@app.command(
    name="rdf",
    help="Create labels for all the objects in the Manifest's resources without labels",
)
def rdf_command(
    manifest: Path = typer.Argument(
        ..., help="The path of the Prez Manifest file to be labelled"
    ),
    context: str = typer.Argument(
        None,
        help="The path of an RDF file, a directory of RDF files or the URL of a SPARQL endpoint from which to obtain labels",
    ),
    username: Annotated[
        str, typer.Option("--username", "-u", help="SPARQL Endpoint username")
    ] = None,
    password: Annotated[
        str, typer.Option("--password", "-p", help="SPARQL Endpoint password")
    ] = None,
) -> None:
    print(
        label(
            manifest,
            LabellerOutputTypes.rdf,
            context,
            make_httpx_client(username, password),
        ).serialize(format="longturtle")
    )


@app.command(
    name="manifest",
    help="Create labels for all the objects in the Manifest's resources without labels and store them as a new Manifest resource",
)
def manifest_command(
    manifest: Path = typer.Argument(
        ..., help="The path of the Prez Manifest file to be labelled"
    ),
    context: Path = typer.Argument(
        ...,
        help="The path of an RDF file, a directory of RDF files or the URL of a SPARQL endpoint from which to obtain labels",
    ),
    username: Annotated[
        str, typer.Option("--username", "-u", help="SPARQL Endpoint username")
    ] = None,
    password: Annotated[
        str, typer.Option("--password", "-p", help="SPARQL Endpoint password")
    ] = None,
) -> None:
    label(
        manifest,
        LabellerOutputTypes.manifest,
        context,
        make_httpx_client(username, password),
    )

    print("A new Resource containing labels has been added to the Manifest")



from pathlib import Path
from textwrap import dedent

import pytest
from kurra.utils import load_graph
from rdflib.compare import isomorphic
from typer.testing import CliRunner

from prezmanifest.cli import app
from prezmanifest.documentor import TableFormats, catalogue, table
from prezmanifest.validator import ManifestValidationError

runner = CliRunner()


def test_create_table_01():
    expected = dedent(
        """
        Resource | Role | Description
        --- | --- | ---
        Catalogue Definition:<br />[`catalogue.ttl`](catalogue.ttl) | [Catalogue Data](https://prez.dev/ManifestResourceRoles/CatalogueData) | The definition of, and medata for, the container which here is a dcat:Catalog object
        Resource Data:<br />[`vocabs/*.ttl`](vocabs/*.ttl) | [Resource Data](https://prez.dev/ManifestResourceRoles/ResourceData) | skos:ConceptScheme objects in RDF (Turtle) files in the vocabs/ folder
        Profile Definition:<br />[`ogc_records_profile.ttl`](https://raw.githubusercontent.com/RDFLib/prez/refs/heads/main/prez/reference_data/profiles/ogc_records_profile.ttl) | [Catalogue & Resource Model](https://prez.dev/ManifestResourceRoles/CatalogueAndResourceModel) | The default Prez profile for Records API
        Labels:<br />[`_background/labels.ttl`](_background/labels.ttl) | [Complete Catalogue and Resource Labels](https://prez.dev/ManifestResourceRoles/CompleteCatalogueAndResourceLabels) | An RDF file containing all the labels for the container content                
        """
    ).strip()

    result = table(Path(__file__).parent / "demo-vocabs" / "manifest.ttl")

    print()
    print()
    print(expected)
    print()
    print()
    print(result)
    print()
    print()

    assert result == expected


def test_create_table_02():
    with pytest.raises(ManifestValidationError):
        table(Path(__file__).parent / "demo-vocabs" / "manifest-invalid-01.ttl")


def test_create_catalogue():
    expected = load_graph(Path(__file__).parent / "demo-vocabs" / "catalogue.ttl")
    actual = catalogue(Path(__file__).parent / "demo-vocabs" / "manifest-cat.ttl")

    assert isomorphic(actual, expected)


def test_create_table_multi():
    expected = dedent(
        """
        Resource | Role | Description
        --- | --- | ---
        Catalogue Definition:<br />[`catalogue-metadata.ttl`](catalogue-metadata.ttl) | [Catalogue Data](https://prez.dev/ManifestResourceRoles/CatalogueData) | The definition of, and medata for, the container which here is a dcat:Catalog object
        Resource Data:<br />[`vocabs/image-test.ttl`](vocabs/image-test.ttl)<br />[`vocabs/language-test.ttl`](vocabs/language-test.ttl) | [Resource Data](https://prez.dev/ManifestResourceRoles/ResourceData) | skos:ConceptScheme objects in RDF (Turtle) files in the vocabs/ folder
        Profile Definition:<br />[`ogc_records_profile.ttl`](https://raw.githubusercontent.com/RDFLib/prez/refs/heads/main/prez/reference_data/profiles/ogc_records_profile.ttl) | [Catalogue & Resource Model](https://prez.dev/ManifestResourceRoles/CatalogueAndResourceModel) | The default Prez profile for Records API
        Labels:<br />[`_background/labels.ttl`](_background/labels.ttl) | [Complete Catalogue and Resource Labels](https://prez.dev/ManifestResourceRoles/CompleteCatalogueAndResourceLabels) | An RDF file containing all the labels for the container content                
        """
    ).strip()

    result = table(Path(__file__).parent / "demo-vocabs" / "manifest-multi.ttl")

    print()
    print()
    print(expected)
    print()
    print()
    print(result)
    print()
    print()

    assert result == expected


def test_create_table_multi_asciidoc():
    expected = dedent(
        """
        |===
        | Resource | Role | Description
        
        | Catalogue Definition: +
         +
        * link:catalogue-metadata.ttl[`catalogue-metadata.ttl`] | https://prez.dev/ManifestResourceRoles/CatalogueData[Catalogue Data] | The definition of, and medata for, the container which here is a dcat:Catalog object
        | Resource Data: +
         +
        * link:vocabs/image-test.ttl[`vocabs/image-test.ttl`] +
        * link:vocabs/language-test.ttl[`vocabs/language-test.ttl`] | https://prez.dev/ManifestResourceRoles/ResourceData[Resource Data] | skos:ConceptScheme objects in RDF (Turtle) files in the vocabs/ folder
        | Profile Definition: +
         +
        * link:https://raw.githubusercontent.com/RDFLib/prez/refs/heads/main/prez/reference_data/profiles/ogc_records_profile.ttl[`ogc_records_profile.ttl`] | https://prez.dev/ManifestResourceRoles/CatalogueAndResourceModel[Catalogue & Resource Model] | The default Prez profile for Records API
        | Labels: +
         +
        * link:_background/labels.ttl[`_background/labels.ttl`] | https://prez.dev/ManifestResourceRoles/CompleteCatalogueAndResourceLabels[Complete Catalogue and Resource Labels] | An RDF file containing all the labels for the container content
        |===  
        """
    ).strip()

    result = table(
        Path(__file__).parent / "demo-vocabs" / "manifest-multi.ttl",
        table_format=TableFormats.asciidoc,
    )

    print()
    print()
    print(expected)
    print()
    print()
    print(result)
    print()
    print()

    assert result == expected


def test_create_table_main_entity():
    expected = dedent(
        """
        Resource | Role | Description
        --- | --- | ---
        Catalogue Definition:<br />[`catalogue.ttl`](catalogue.ttl) | [Catalogue Data](https://prez.dev/ManifestResourceRoles/CatalogueData) | The definition of, and medata for, the container which here is a dcat:Catalog object
        Resource Data:<br />[`vocabs/image-test.ttl`](vocabs/image-test.ttl)<br />[`vocabs/language-test.ttl`](vocabs/language-test.ttl) | [Resource Data](https://prez.dev/ManifestResourceRoles/ResourceData) | skos:ConceptScheme objects in RDF (Turtle) files in the vocabs/ folder
        Profile Definition:<br />[`ogc_records_profile.ttl`](https://raw.githubusercontent.com/RDFLib/prez/refs/heads/main/prez/reference_data/profiles/ogc_records_profile.ttl) | [Catalogue & Resource Model](https://prez.dev/ManifestResourceRoles/CatalogueAndResourceModel) | The default Prez profile for Records API
        Labels:<br />[`_background/labels.ttl`](_background/labels.ttl) | [Complete Catalogue and Resource Labels](https://prez.dev/ManifestResourceRoles/CompleteCatalogueAndResourceLabels) | An RDF file containing all the labels for the container content                
        """
    ).strip()

    result = table(Path(__file__).parent / "demo-vocabs" / "manifest-mainEntity.ttl")

    print()
    print()
    print(expected)
    print()
    print()
    print(result)
    print()
    print()

    assert result == expected


def test_create_catalogue_multi():
    expected = load_graph(Path(__file__).parent / "demo-vocabs" / "catalogue.ttl")
    actual = catalogue(Path(__file__).parent / "demo-vocabs" / "manifest-multi.ttl")

    assert isomorphic(actual, expected)


def test_create_catalogue_main_entity():
    expected = load_graph(Path(__file__).parent / "demo-vocabs" / "catalogue.ttl")
    actual = catalogue(
        Path(__file__).parent / "demo-vocabs" / "manifest-mainEntity.ttl"
    )

    assert isomorphic(actual, expected)


def test_table_cli():
    result = runner.invoke(
        app,
        [
            "document",
            "table",
            str(Path(__file__).parent / "demo-vocabs" / "manifest.ttl"),
        ],
    )
    actual = result.stdout

    assert "Catalogue Definition" in actual


def test_catalogue_cli():
    expected = load_graph(
        """
        PREFIX ns1: <http://purl.org/linked-data/registry#>
        PREFIX schema: <https://schema.org/>
        PREFIX xsd: <http://www.w3.org/2001/XMLSchema#>
        
        <https://example.com/demo-vocabs>
            a schema:DataCatalog ;
            ns1:status <https://linked.data.gov.au/def/reg-statuses/experimental> ;
            schema:codeRepository "https://github.com/kurrawong/demo-vocabs" ;
            schema:creator <https://kurrawong.ai> ;
            schema:dateCreated "2023"^^xsd:gYear ;
            schema:dateModified "2024-10-16"^^xsd:date ;
            schema:description "A testing catalogue for the Prez Manifest Loader tool" ;
            schema:hasPart
                <https://example.com/demo-vocabs/image-test> ,
                <https://example.com/demo-vocabs/language-test> ;
            schema:name "Demo Vocabularies" ;
            schema:publisher <https://kurrawong.ai> ;
            schema:version "1.0.1" ;
        .        
        """
    )
    result = runner.invoke(
        app,
        [
            "document",
            "catalogue",
            str(Path(__file__).parent / "demo-vocabs" / "manifest.ttl"),
        ],
    )
    actual = load_graph(result.stdout)

    assert isomorphic(expected, actual)



"""
This script creates a "Prez Resources" table in either Markdown or ASCIIDOC from a Manifest file which it validates first

Run this script with the -h flag for more help, i.e. ~$ python documentor.py -h

Example:

Input:

PREFIX mrr: <https://prez.dev/ManifestResourceRoles/>
PREFIX prez: <https://prez.dev/>
PREFIX prof: <http://www.w3.org/ns/dx/prof/>
PREFIX schema: <https://schema.org/>

[]
    a prez:Manifest ;
    prof:hasResource
        [
            prof:hasArtifact "catalogue.ttl" ;
            prof:hasRole mrr:ContainerData ;
            schema:description "The definition of, and medata for, the container which here is a dcat:Catalog object" ;
            schema:name "Catalogue Definition"
        ] ,
        [
            prof:hasArtifact "vocabs/*.ttl" ;
            prof:hasRole mrr:ContentData ;
            schema:description "skos:ConceptScheme objects in RDF (Turtle) files in the vocabs/ folder" ;
            schema:name "Content"
        ] ,
        [
            prof:hasArtifact "https://github.com/RDFLib/prez/blob/main/prez/reference_data/profiles/ogc_records_profile.ttl" ;
            prof:hasRole mrr:ContainerAndContentModel ;
            schema:description "The default Prez profile for Records API" ;
            schema:name "Profile Definition"
        ] ,
        [
            prof:hasArtifact "_background/labels.ttl" ;
            prof:hasRole mrr:CompleteContainerAndContentLabels ;
            schema:description "An RDF file containing all the labels for the container content" ;
        ] ;
    .

Output:

Resource | Role | Description
--- | --- | ---
Catalogue Definition, [`catalogue.ttl`](catalogue.ttl) | [Container Data](https://prez.dev/ManifestResourceRoles/ContainerData) | The definition of, and medata for, the container which here is a dcat:Catalog object
Content, [`vocabs/*.ttl`](vocabs/*.ttl) | [Content Data](https://prez.dev/ManifestResourceRoles/ContentData) | skos:ConceptScheme objects in RDF (Turtle) files in the vocabs/ folder
Profile Definition, [`ogc_records_profile.ttl`](https://github.com/RDFLib/prez/blob/main/prez/reference_data/profiles/ogc_records_profile.ttl) | [Container & Content Model](https://prez.dev/ManifestResourceRoles/ContainerAndContentModel) | The default Prez profile for Records API
Labels file, [`_background/labels.ttl`](_background/labels.ttl) | [Complete Content and Container Labels](https://prez.dev/ManifestResourceRoles/CompleteContainerAndContentLabels) | An RDF file containing all the labels for the container content
"""

from enum import Enum
from pathlib import Path

from rdflib import Graph, Literal, URIRef
from rdflib.namespace import DCAT, PROF, RDF, SDO, SKOS

from prezmanifest.definednamespaces import MRR
from prezmanifest.utils import (
    get_files_from_artifact,
    get_identifier_from_file,
    load_graph,
)
from prezmanifest.utils import get_manifest_paths_and_graph
from prezmanifest.validator import validate


class TableFormats(str, Enum):
    asciidoc = "asciidoc"
    markdown = "markdown"


def table(manifest: Path, table_format: TableFormats = TableFormats.markdown) -> str:
    if not isinstance(table_format, TableFormats):
        raise ValueError(
            f"Invalid table_format value, must be one of {', '.join([x for x in TableFormats])}"
        )

    # load and validate manifest
    validate(manifest)
    manifest_graph = load_graph(manifest)

    # add in MRR vocab
    manifest_graph += load_graph(Path(__file__).parent / "mrr.ttl")

    if table_format == TableFormats.asciidoc:
        header = "|===\n| Resource | Role | Description\n\n"
    else:
        header = "Resource | Role | Description\n--- | --- | ---\n"

    body = ""
    for s, o in manifest_graph.subject_objects(PROF.hasResource):
        artifact_docs = []
        for artifact in manifest_graph.objects(o, PROF.hasArtifact):
            if isinstance(artifact, Literal):
                a = str(artifact)
            else:  # isinstance( a, BNode)
                a = str(
                    manifest_graph.value(
                        subject=artifact, predicate=SDO.contentLocation
                    )
                )
            if table_format == TableFormats.asciidoc:
                artifact_docs.append(
                    f"link:{a}[`{a.split('/')[-1] if a.startswith('http') else a}`]"
                )
            else:
                artifact_docs.append(
                    f"[`{a.split('/')[-1] if a.startswith('http') else a}`]({a})"
                )
        role_iri = manifest_graph.value(o, PROF.hasRole)
        role_label = manifest_graph.value(role_iri, SKOS.prefLabel)
        if table_format == TableFormats.asciidoc:
            role = f"{role_iri}[{role_label}]"
        else:
            role = f"[{role_label}]({role_iri})"
        name = manifest_graph.value(o, SDO.name)
        description = manifest_graph.value(o, SDO.description)
        if table_format == TableFormats.asciidoc:
            n = (
                f"""{name}: +
 +
* {
                    ''' +
* '''.join(artifact_docs)
                }"""
                if name is not None
                else f"{
                    ''' +
* '''.join(artifact_docs)
                }"
                ""
            )
        else:
            n = (
                f"{name}:<br />{'<br />'.join(artifact_docs)}"
                if name is not None
                else f"{'<br />'.join(artifact_docs)}"
            )
        d = description if description is not None else ""
        if table_format == TableFormats.asciidoc:
            body += f"| {n} | {role} | {d}\n"
        else:
            body += f"{n} | {role} | {d}\n"

    if table_format == TableFormats.asciidoc:
        footer = "|===\n"
    else:
        footer = ""

    return (header + body + footer).strip()


def catalogue(manifest: Path) -> Graph:
    manifest_path, manifest_root, manifest_graph = get_manifest_paths_and_graph(
        manifest
    )

    catalogue = Graph()
    for s, o in manifest_graph.subject_objects(PROF.hasResource):
        for role in manifest_graph.objects(o, PROF.hasRole):
            if role == MRR.CatalogueData:
                for artifact in manifest_graph.objects(o, PROF.hasArtifact):
                    # the artifact can only be a triples file (not a quads file)
                    catalogue = load_graph(manifest_root / artifact)

    # get the IRI of the catalogue
    catalogue_iri = catalogue.value(
        predicate=RDF.type, object=DCAT.Catalog
    ) or catalogue.value(predicate=RDF.type, object=SDO.DataCatalog)

    # non-catalogue resources
    for s, o in manifest_graph.subject_objects(PROF.hasResource):
        for role in manifest_graph.objects(o, PROF.hasRole):
            if role == MRR.ResourceData:
                for artifact in manifest_graph.objects(o, PROF.hasArtifact):
                    for f in get_files_from_artifact(
                        (manifest_path, manifest_root, manifest_graph), artifact
                    ):
                        if isinstance(artifact, Literal):
                            for iri in sorted(get_identifier_from_file(f)):
                                if iri != URIRef("urn:x-rdflib:default"):
                                    catalogue.add((catalogue_iri, SDO.hasPart, iri))
                        else:  # isinstance(artifact, BNode):
                            iri = manifest_graph.value(
                                subject=artifact, predicate=SDO.mainEntity
                            )
                            catalogue.add((catalogue_iri, SDO.hasPart, iri))

    return catalogue



from datetime import datetime

import pytest
from kurra.db import upload
from typer.testing import CliRunner

import prezmanifest.loader
from prezmanifest.utils import *
from tests.fuseki.conftest import fuseki_container

runner = CliRunner()
import httpx

TESTS_DIR = Path(__file__).resolve().parent


def test_path_or_url():
    s1 = "http://example.com"
    assert path_or_url(s1) == s1

    s2 = "/usr/local"
    assert path_or_url(s2) == Path(s2)


def test_localise_path():
    p = Path("/usr/lib/some-file.txt")
    r = Path("/usr/lib")

    x = localise_path(p, r)

    assert x == Path("some-file.txt")

    p = "/usr/lib/some-file.txt"
    r = Path("/usr/lib")

    x = localise_path(p, r)

    assert x == Path("some-file.txt")

    p = "http://example.com/file/one.txt"
    r = Path("/usr/lib")

    x = localise_path(p, r)

    assert x == p


def test_absolutise_path():
    p = Path("some-file.txt")
    r = Path("/usr/lib")

    x = absolutise_path(p, r)

    assert x == Path("/usr/lib/some-file.txt")

    p = "http://example.com/file/one.txt"
    r = Path("/usr/lib")

    x = absolutise_path(p, r)

    assert x == p


def test_get_files_from_artifact():
    MANIFEST = TESTS_DIR / "demo-vocabs" / "manifest.ttl"

    fs = list(get_files_from_artifact(MANIFEST, Literal("vocabs/*.ttl")))

    assert len(fs) == 2
    assert (
        Path(
            "/Users/nick/work/kurrawong/prez-manifest/tests/demo-vocabs/vocabs/image-test.ttl"
        )
        in fs
    )
    assert (
        Path(
            "/Users/nick/work/kurrawong/prez-manifest/tests/demo-vocabs/vocabs/language-test.ttl"
        )
        in fs
    )


def test_get_identifier_from_file():
    f1 = TESTS_DIR / "demo-vocabs" / "vocabs" / "image-test.ttl"

    i = get_identifier_from_file(f1)
    assert i[0] == URIRef("https://example.com/demo-vocabs/image-test")


def test_get_validator_graph():
    MANIFEST = TESTS_DIR / "demo-vocabs" / "manifest.ttl"

    g = get_validator_graph(MANIFEST, URIRef("https://data.idnau.org/pid/cp"))

    assert len(g) == 318

    g2 = get_validator_graph(
        MANIFEST, TESTS_DIR / "demo-vocabs" / "vocabs" / "image-test.ttl"
    )

    assert len(g2) == 29


def test_get_manifest_paths_and_graph():
    MANIFEST = TESTS_DIR / "demo-vocabs" / "manifest.ttl"

    extracted_file_path, manifest_root, manifest_graph = get_manifest_paths_and_graph(
        MANIFEST
    )

    assert extracted_file_path == MANIFEST
    assert manifest_root == MANIFEST.parent
    assert len(manifest_graph) == 21


def test_get_catalogue_iri_from_manifest():
    MANIFEST = TESTS_DIR / "demo-vocabs" / "manifest.ttl"

    assert get_catalogue_iri_from_manifest(MANIFEST) == URIRef(
        "https://example.com/demo-vocabs"
    )


def test_target_contains_this_manifests_catalogue(fuseki_container):
    MANIFEST = TESTS_DIR / "demo-vocabs" / "manifest.ttl"
    port = fuseki_container.get_exposed_port(3030)

    with httpx.Client() as client:
        SPARQL_ENDPOINT = f"http://localhost:{port}/ds"

        # positive test
        sparql(SPARQL_ENDPOINT, "DROP ALL", client)
        prezmanifest.loader.load(MANIFEST, SPARQL_ENDPOINT)
        assert target_contains_this_manifests_catalogue(MANIFEST, SPARQL_ENDPOINT)

        # negative test
        sparql(SPARQL_ENDPOINT, "DROP ALL", client)
        assert not target_contains_this_manifests_catalogue(MANIFEST, SPARQL_ENDPOINT)


# TODO add tests
def test_make_httpx_client():
    pass


def test_get_main_entity_iri_of_artifact():
    MANIFEST = TESTS_DIR / "demo-vocabs" / "manifest-conformance.ttl"

    assert get_main_entity_iri_of_artifact(
        MANIFEST.parent / "vocabs/image-test.ttl", MANIFEST
    ) == URIRef("https://example.com/demo-vocabs/image-test")

    assert get_main_entity_iri_of_artifact(
        MANIFEST.parent / "vocabs/language-test.ttl", MANIFEST
    ) == URIRef("https://example.com/demo-vocabs/language-test")


def test_get_version_indicators_local():
    MANIFEST = TESTS_DIR / "demo-vocabs" / "manifest.ttl"

    vi = {}
    get_version_indicators_local(
        MANIFEST, TESTS_DIR / "demo-vocabs" / "vocabs" / "language-test.ttl", vi
    )

    assert vi["modified_date"] == date_parse("2024-11-21")
    assert vi["version_iri"] == "https://example.com/demo-vocabs/language-test/1.0"

    with pytest.raises(ValueError):
        vi = {}
        get_version_indicators_local(
            MANIFEST, TESTS_DIR / "demo-vocabs" / "vocabs" / "language-testx.ttl", vi
        )


def test_get_version_indicators_sparql(fuseki_container):
    port = fuseki_container.get_exposed_port(3030)
    SPARQL_ENDPOINT = f"http://localhost:{port}/ds"
    ASSET_PATH = TESTS_DIR / "demo-vocabs" / "vocabs" / "language-test.ttl"
    ASSET_GRAPH_IRI = "https://example.com/demo-vocabs/language-test"

    c = make_httpx_client()

    upload(SPARQL_ENDPOINT, ASSET_PATH, ASSET_GRAPH_IRI, False, http_client=c)

    q = """
        SELECT (COUNT(*) AS ?count) 
        WHERE {
          GRAPH <XXX> {
            ?s ?p ?o
          }
        }        
        """.replace(
        "XXX", ASSET_GRAPH_IRI
    )

    r = query(SPARQL_ENDPOINT, q, c, return_python=True, return_bindings_only=True)

    count = int(r[0]["count"]["value"])

    assert count == 71

    r = get_version_indicators_sparql(ASSET_GRAPH_IRI, SPARQL_ENDPOINT, http_client=c)

    assert r["modified_date"] == date_parse("2024-11-21")


def test_compare_version_indicators():
    one = {
        "modified_date": date_parse("2024-11-20"),
        "version_info": None,
        "version_iri": None,
        "file_size": None,
        "main_entity_iri": None,
    }

    two = {
        "modified_date": date_parse("2024-11-21"),
        "version_info": "1.1",
        "version_iri": None,
        "file_size": None,
        "main_entity_iri": None,
    }

    assert compare_version_indicators(one, two) == VersionIndicatorComparison.Second

    three = {
        "modified_date": None,
        "version_info": "1.2.2",
        "version_iri": None,
        "file_size": None,
        "main_entity_iri": None,
    }

    assert compare_version_indicators(two, three) == VersionIndicatorComparison.Second

    four = {
        "modified_date": None,
        "version_info": "1.2.3",
        "version_iri": None,
        "file_size": None,
        "main_entity_iri": None,
    }

    assert compare_version_indicators(three, four) == VersionIndicatorComparison.Second

    five = {
        "modified_date": None,
        "version_info": None,
        "version_iri": "https://example.com/demo-vocabs/language-test/1",
        "file_size": None,
        "main_entity_iri": None,
    }

    six = {
        "modified_date": None,
        "version_info": None,
        "version_iri": "https://example.com/demo-vocabs/language-test/2.0",
        "file_size": None,
        "main_entity_iri": None,
    }

    assert compare_version_indicators(five, six) == VersionIndicatorComparison.Second

    seven = {
        "modified_date": None,
        "version_info": None,
        "version_iri": "https://example.com/demo-vocabs/language-test/2.1",
        "file_size": None,
        "main_entity_iri": None,
    }

    assert compare_version_indicators(six, seven) == VersionIndicatorComparison.Second


def test_which_is_more_recent(fuseki_container):
    port = fuseki_container.get_exposed_port(3030)
    SPARQL_ENDPOINT = f"http://localhost:{port}/ds"
    ARTIFACT_PATH = TESTS_DIR / "demo-vocabs" / "vocabs" / "language-test.ttl"
    ARTIFACT_MAIN_ENTITY = "https://example.com/demo-vocabs/language-test"

    c = make_httpx_client()

    upload(SPARQL_ENDPOINT, ARTIFACT_PATH, ARTIFACT_MAIN_ENTITY, False, c)

    vi = {
        "main_entity": URIRef("https://example.com/demo-vocabs/language-test"),
        "modified_date": datetime(2025, 2, 28, 0, 0),
        "version_iri": "https://example.com/demo-vocabs/language-test/1.1",
        "file_size": 5053,
    }

    r = which_is_more_recent(vi, SPARQL_ENDPOINT, c)

    assert r == VersionIndicatorComparison.First

    vi2 = {
        "main_entity": URIRef("https://example.com/demo-vocabs/language-test"),
        "modified_date": datetime(2023, 2, 28, 0, 0),
        "file_size": 5053,
    }

    r = which_is_more_recent(vi2, SPARQL_ENDPOINT, c)

    assert r == VersionIndicatorComparison.Second


def test_denormalise_artifacts():
    m = TESTS_DIR / "demo-vocabs" / "manifest-conformance.ttl"
    manifest_path, manifest_root, manifest_graph = get_manifest_paths_and_graph(m)
    x = denormalise_artifacts((manifest_path, manifest_root, manifest_graph))
    assert manifest_root / "catalogue.ttl" in x.keys()
    assert manifest_root / "vocabs/image-test.ttl" in x.keys()
    assert manifest_root / "vocabs/language-test.ttl" in x.keys()
    assert (
        "https://raw.githubusercontent.com/RDFLib/prez/refs/heads/main/prez/reference_data/profiles/ogc_records_profile.ttl"
        in x.keys()
    )
    assert manifest_root / "_background/labels.ttl" in x.keys()

    assert x[manifest_root / "vocabs/image-test.ttl"]["file_size"] == 20324

    m = TESTS_DIR / "demo-vocabs" / "manifest-conformance-all.ttl"
    manifest_path, manifest_root, manifest_graph = get_manifest_paths_and_graph(m)
    x = denormalise_artifacts(m)
    assert manifest_root / "catalogue.ttl" in x.keys()
    assert manifest_root / "vocabs/image-test.ttl" in x.keys()
    assert manifest_root / "vocabs/language-test.ttl" in x.keys()
    assert (
        "https://raw.githubusercontent.com/RDFLib/prez/refs/heads/main/prez/reference_data/profiles/ogc_records_profile.ttl"
        in x.keys()
    )
    assert manifest_root / "_background/labels.ttl" in x.keys()



import collections
import json
from pathlib import Path
from typing import Annotated

import typer
from rich.table import Table

from prezmanifest.cli.app import app
from prezmanifest.cli.console import console
from prezmanifest.syncer import sync
from prezmanifest.utils import make_httpx_client


@app.command(
    name="sync",
    help="Synchronize a Prez Manifest's resources with loaded copies of them in a SPARQL Endpoint",
)
def sync_command(
    manifest: Path = typer.Argument(
        ..., help="The path of the Prez Manifest file to be loaded"
    ),
    endpoint: str = typer.Argument(..., help="The URL of the SPARQL Endpoint"),
    update_remote: bool = typer.Argument(
        True, help="Copy more recent local artifacts to DB"
    ),
    update_local: bool = typer.Argument(
        True, help="Copy more recent DB artifacts to local"
    ),
    add_remote: bool = typer.Argument(True, help="Add new local artifacts to DB"),
    add_local: bool = typer.Argument(True, help="Add new DB artifacts to local"),
    username: Annotated[
        str, typer.Option("--username", "-u", help="SPARQL Endpoint username")
    ] = None,
    password: Annotated[
        str, typer.Option("--password", "-p", help="SPARQL Endpoint password")
    ] = None,
    response_format: str = typer.Option(
        "table",
        "--response-format",
        "-f",
        help="The response format of the SPARQL query. Either 'table' (default) or 'json'",
    ),
) -> None:
    r = sync(
        manifest,
        endpoint,
        make_httpx_client(username, password),
        update_remote,
        update_local,
        add_remote,
        add_local,
    )

    if response_format == "json":
        print(json.dumps(r, indent=4))
    else:
        console.print(result_as_rich_table(r))


def result_as_rich_table(sync_status: dict):
    t = Table()
    t.add_column("Artifact")
    t.add_column("Main Entity")
    t.add_column("Direction")

    for k, v in collections.OrderedDict(sorted(sync_status.items())).items():
        t.add_row(str(k), str(v["main_entity"]), v["direction"])

    # json.dumps(sync_status, indent=4)

    return t



from pathlib import Path

import typer

from prezmanifest.cli.app import app
from prezmanifest.validator import validate


@app.command(
    name="validate", help="Validate the structure and content of a Prez Manifest"
)
def validate_command(
    manifest: Path = typer.Argument(
        ..., help="The path of the Prez Manifest file to be validated"
    ),
) -> None:
    validate(manifest)



from pathlib import Path
from typing import Annotated

import typer

from prezmanifest.loader import load

app = typer.Typer(help="Load a Prez Manifest's content into a file or DB")


@app.command(
    name="sparql", help="Load a Prez Manifest's resources into a SPARQL Endpoint"
)
def sparql_command(
    manifest: Path = typer.Argument(
        ..., help="The path of the Prez Manifest file to be loaded"
    ),
    endpoint: str = typer.Argument(..., help="The URL of the SPARQL Endpoint"),
    username: Annotated[
        str, typer.Option("--username", "-u", help="SPARQL Endpoint username.")
    ] = None,
    password: Annotated[
        str, typer.Option("--password", "-p", help="SPARQL Endpoint password.")
    ] = None,
) -> None:
    load(
        manifest,
        sparql_endpoint=endpoint,
        sparql_username=username,
        sparql_password=password,
    )


@app.command(
    name="file", help="Load a Prez Manifest's resources into a single RDF quads file"
)
def file_command(
    manifest: Path = typer.Argument(
        ..., help="The path of the Prez Manifest file to be loaded"
    ),
    file: Path = typer.Argument(..., help="The path of the quads file"),
) -> None:
    load(manifest, destination_file=file)



from pathlib import Path

import pytest
from testcontainers.core.container import DockerContainer
from testcontainers.core.waiting_utils import wait_for_logs

FUSEKI_IMAGE = "ghcr.io/kurrawong/fuseki-geosparql:git-main-e642d849"


@pytest.fixture(scope="module")
def fuseki_container(request: pytest.FixtureRequest):
    container = DockerContainer(FUSEKI_IMAGE)
    container.with_volume_mapping(
        str(Path(__file__).parent / "shiro.ini"), "/opt/fuseki/shiro.ini"
    )
    container.with_volume_mapping(
        str(Path(__file__).parent / "config.ttl"), "/fuseki/config.ttl"
    )
    container.with_exposed_ports(3030)
    container.start()
    wait_for_logs(container, "Started")

    def cleanup():
        container.stop()

    request.addfinalizer(cleanup)
    return container



"""
Assesses a given Manifest, finds any IRIs in any of the given resources missing labels and tries to patch them from
a given source of labels, such as KurrawongAI's Semantic Background (https://github.com/kurrawong/semantic-background)
repository.
"""

from enum import Enum
from pathlib import Path

import httpx
from kurra.utils import load_graph
from labelify import extract_labels, find_missing_labels
from rdflib import BNode, Graph, Literal
from rdflib.namespace import PROF, RDF

from prezmanifest.definednamespaces import MRR, PREZ
from prezmanifest.utils import get_manifest_paths_and_graph, denormalise_artifacts


class LabellerOutputTypes(str, Enum):
    iris = "iris"
    rdf = "rdf"
    manifest = "manifest"


def label(
    manifest: Path,
    output_type: LabellerOutputTypes = LabellerOutputTypes.manifest,
    additional_context: Path | str | Graph = None,
    http_client: httpx.Client = None,
) -> set | Graph | None:
    """ "Main function for labeller module"""
    if not isinstance(output_type, LabellerOutputTypes):
        raise ValueError(
            f"Invalid output_type value, must be one of {', '.join([x for x in LabellerOutputTypes])}"
        )

    # create the target from the Manifest
    manifest_path, manifest_root, manifest_graph = get_manifest_paths_and_graph(
        manifest
    )

    content_graph = Graph()
    context_graph = Graph()

    artifacts = denormalise_artifacts((manifest_path, manifest_root, manifest_graph))

    for k, v in artifacts.items():
        if v["role"] in [MRR.CatalogueData, MRR.ResourceData]:
            content_graph += load_graph(k)
        elif v["role"] in [
            MRR.CompleteCatalogueAndResourceLabels,
            MRR.IncompleteCatalogueAndResourceLabels,
        ]:
            context_graph += load_graph(k)

    # add labels for system IRIs
    context_graph.parse(Path(__file__).parent / "system-labels.ttl")

    if output_type == LabellerOutputTypes.iris:
        combined_graph = manifest_graph + content_graph + context_graph

        iris_missing_labels = find_missing_labels(
            combined_graph, additional_context, http_client=http_client
        )

        return iris_missing_labels

    elif output_type == LabellerOutputTypes.rdf:
        if additional_context is None:
            raise ValueError("You must provide additional context")

        combined_graph = manifest_graph + content_graph + context_graph

        iris_missing_labels = find_missing_labels(
            combined_graph, http_client=http_client
        )

        return extract_labels(iris_missing_labels, additional_context, http_client)

    else:  # output_type == LabellerOutputTypes.manifest
        # If this is selected, generate the "rdf" output and create a resource for it in the Manifest
        # If there are no more missing labels then we have a mrr:CompleteCatalogueAndResourceLabels
        # else add mrr:IncompleteCatalogueAndResourceLabels

        # Generate labels for any IRIs missing them, using context given in the Manifest and any
        # Additional Context supplied

        rdf_addition = label(manifest, LabellerOutputTypes.rdf, additional_context)

        if len(rdf_addition) > 0:
            new_artifact = manifest.parent / "labels-additional.ttl"
            rdf_addition.serialize(destination=new_artifact, format="longturtle")
            new_resource = BNode()

            # Add to the Manifest
            manifest_iri = manifest_graph.value(
                predicate=RDF.type, object=PREZ.Manifest
            )
            manifest_graph.add((manifest_iri, PROF.hasResource, new_resource))
            manifest_graph.add(
                (new_resource, PROF.hasRole, MRR.IncompleteCatalogueAndResourceLabels)
            )
            manifest_graph.add(
                (new_resource, PROF.hasArtifact, Literal(new_artifact.name))
            )

            manifest_graph.serialize(destination=manifest, format="longturtle")
        else:
            raise Warning(
                "No new labels have been generated for content in this Manifest. "
                "This could be because none were missing or because no new labels can be found in any "
                "supplied additional context."
            )



from pathlib import Path

import typer

from prezmanifest.documentor import TableFormats, catalogue, table

app = typer.Typer(help="Create documentation from a Prez Manifest")


@app.command(
    name="table",
    help="Create a Markdown or ASCIIDOC table for the resources listed in a Prez Manifest",
)
def table_command(
    manifest: Path = typer.Argument(
        ..., help="The path of the Prez Manifest file to be documented"
    ),
    table_format: TableFormats = typer.Option(
        TableFormats.markdown,
        "--format",
        "-f",
        help="The format of the table to be created",
    ),
) -> None:
    print(table(manifest, table_format))


@app.command(
    name="catalogue",
    help="Add the resources listed in a Prez Manifest to a catalogue RDF file",
)
def catalogue_command(
    manifest: Path = typer.Argument(
        ..., help="The path of the Prez Manifest file to be documented"
    ),
) -> None:
    print(catalogue(manifest).serialize(format="longturtle"))



import warnings
from pathlib import Path

import httpx
import pytest
from kurra.db import sparql, upload
from rdflib import Dataset, URIRef
from typer.testing import CliRunner

from prezmanifest.loader import ReturnDatatype, load
from tests.fuseki.conftest import fuseki_container

runner = CliRunner()


def test_load_only_one_set():
    warnings.filterwarnings(
        "ignore", category=DeprecationWarning
    )  # ignore RDFLib's ConjunctiveGraph warning

    manifest = Path(Path(__file__).parent / "demo-vocabs/manifest.ttl")

    with pytest.raises(ValueError):
        load(manifest)

    with pytest.raises(ValueError):
        load(
            manifest,
            sparql_endpoint="http://fake.com",
            destination_file=Path("some-fake-path"),
        )

    with pytest.raises(ValueError):
        load(
            manifest,
            destination_file=Path("some-fake-path"),
            return_data_type=ReturnDatatype.graph,
        )

    with pytest.raises(ValueError):
        load(manifest, return_data_type="hello")

    load(manifest, destination_file=Path("temp.trig"))

    Path("temp.trig").unlink(missing_ok=True)


def test_fuseki_query(fuseki_container):
    port = fuseki_container.get_exposed_port(3030)
    SPARQL_ENDPOINT = f"http://localhost:{port}/ds"
    TESTING_GRAPH = "https://example.com/testing-graph"

    data = """
            PREFIX ex: <http://example.com/>

            ex:a ex:b ex:c .
            ex:a2 ex:b2 ex:c2 .
            """

    upload(SPARQL_ENDPOINT, data, TESTING_GRAPH, False)

    q = """
        SELECT (COUNT(*) AS ?count) 
        WHERE {
          GRAPH <XXX> {
            ?s ?p ?o
          }
        }        
        """.replace(
        "XXX", TESTING_GRAPH
    )

    r = sparql(SPARQL_ENDPOINT, q, return_python=True, return_bindings_only=True)

    count = int(r[0]["count"]["value"])

    assert count == 2

    q = "DROP GRAPH <XXX>".replace("XXX", TESTING_GRAPH)

    print("QUERY")
    print(q)
    print("QUERY")

    r = sparql(SPARQL_ENDPOINT, q)

    q = """
        SELECT (COUNT(*) AS ?count) 
        WHERE {
          GRAPH <XXX> {
            ?s ?p ?o
          }
        }        
        """.replace(
        "XXX", TESTING_GRAPH
    )

    r = sparql(SPARQL_ENDPOINT, q, return_python=True, return_bindings_only=True)

    count = int(r[0]["count"]["value"])

    assert count == 0


def test_load_to_quads_file():
    warnings.filterwarnings(
        "ignore", category=DeprecationWarning
    )  # ignore RDFLib's ConjunctiveGraph warning
    manifest = Path(__file__).parent / "demo-vocabs" / "manifest.ttl"
    results_file = Path(__file__).parent / "results.trig"

    # extract all Manifest content into an n-quads file
    load(manifest, sparql_endpoint=None, destination_file=results_file)

    # load the resultant Dataset to test it
    d = Dataset()
    d.parse(results_file, format="trig")

    # get a list of IDs of the Graphs in the Dataset
    graph_ids = [x.identifier for x in d.graphs()]

    # check that each Manifest part has a graph present
    assert URIRef("https://example.com/demo-vocabs-catalogue") in graph_ids
    assert URIRef("https://example.com/demo-vocabs/image-test") in graph_ids
    assert URIRef("https://example.com/demo-vocabs/language-test") in graph_ids
    assert URIRef("http://background") in graph_ids
    assert URIRef("https://olis.dev/SystemGraph") in graph_ids

    Path(results_file).unlink()


def test_load_to_fuseki(fuseki_container):
    SPARQL_ENDPOINT = f"http://localhost:{fuseki_container.get_exposed_port(3030)}/ds"

    manifest = Path(__file__).parent / "demo-vocabs" / "manifest.ttl"
    load(manifest, sparql_endpoint=SPARQL_ENDPOINT)

    q = """
        SELECT (COUNT(DISTINCT ?g) AS ?count)
        WHERE {
            GRAPH ?g {
                ?s ?p ?o 
            }
        }      
        """

    r = sparql(SPARQL_ENDPOINT, q, return_python=True, return_bindings_only=True)

    count = int(r[0]["count"]["value"])

    assert count == 5


def test_load_to_fuseki_basic_auth(fuseki_container):
    SPARQL_ENDPOINT = (
        f"http://localhost:{fuseki_container.get_exposed_port(3030)}/authds"
    )

    manifest = Path(__file__).parent / "demo-vocabs" / "manifest.ttl"
    load(
        manifest,
        sparql_endpoint=SPARQL_ENDPOINT,
        sparql_username="admin",
        sparql_password="admin",
    )

    q = """
        SELECT (COUNT(DISTINCT ?g) AS ?count)
        WHERE {
            GRAPH ?g {
                ?s ?p ?o 
            }
        }      
        """
    client = httpx.Client(auth=("admin", "admin"))
    r = sparql(
        SPARQL_ENDPOINT,
        q,
        return_python=True,
        return_bindings_only=True,
        http_client=client,
    )

    count = int(r[0]["count"]["value"])

    assert count == 5


def test_load_with_artifact_bn():
    warnings.filterwarnings(
        "ignore", category=DeprecationWarning
    )  # ignore RDFLib's ConjunctiveGraph warning
    manifest = Path(__file__).parent / "demo-vocabs" / "manifest-mainEntity.ttl"
    results_file = Path(__file__).parent / "results.trig"

    # extract all Manifest content into an n-quads file
    load(manifest, destination_file=results_file)

    # load the resultant Dataset to test it
    d = Dataset()
    d.parse(results_file, format="trig")

    # get a list of IDs of the Graphs in the Dataset
    graph_ids = [x.identifier for x in d.graphs()]

    # check that each Manifest part has a graph present
    assert URIRef("https://example.com/demo-vocabs-catalogue") in graph_ids
    assert URIRef("https://example.com/demo-vocabs/image-test") in graph_ids
    assert URIRef("https://example.com/demo-vocabs/language-test") in graph_ids
    assert URIRef("http://background") in graph_ids
    assert URIRef("https://olis.dev/SystemGraph") in graph_ids

    Path(results_file).unlink()


# TODO: not working
# def test_load_cli_file(fs):
#     warnings.filterwarnings(
#         "ignore", category=DeprecationWarning
#     )  # ignore RDFLib's ConjunctiveGraph warning
#
#     fake_file = fs.create_file(Path(__file__).parent.resolve() / "temp.trig")
#
#     manifest = Path(__file__).parent / "demo-vocabs/manifest.ttl"
#     tmp_output_file = Path(__file__).parent.resolve() / "temp.trig"
#     runner.invoke(
#         app,
#         [
#             "load",
#             "file",
#             manifest,
#             fake_file.path
#         ],
#     )
#
#     output = fake_file.read_text()
#
#     assert output.count(" {") == 5
#
#     # Path("temp.trig").unlink(missing_ok=True)


# TODO: not working
# def test_load_cli_sparql(fuseki_container):
#     warnings.filterwarnings(
#         "ignore", category=DeprecationWarning
#     )  # ignore RDFLib's ConjunctiveGraph warning
#
#     manifest = Path(__file__).parent / "demo-vocabs/manifest.ttl"
#     SPARQL_ENDPOINT = (
#         f"http://localhost:{fuseki_container.get_exposed_port(3030)}/authds"
#     )
#
#     response = runner.invoke(
#         app,
#         [
#             "load",
#             "sparql",
#             manifest,
#             SPARQL_ENDPOINT,
#             "-u",
#             "admin",
#             "-p",
#             "admin"
#         ],
#     )
#
#     print(response.stdout)
#
#     q = """
#         SELECT (COUNT(DISTINCT ?g) AS ?count)
#         WHERE {
#             GRAPH ?g {
#                 ?s ?p ?o
#             }
#         }
#         """
#     client = httpx.Client(auth=("admin", "admin"))
#     r = sparql(
#         SPARQL_ENDPOINT,
#         q,
#         return_python=True,
#         return_bindings_only=True,
#         http_client=client,
#     )
#
#     count = int(r[0]["count"]["value"])
#
#     assert count == 5



import importlib.metadata

__version__ = importlib.metadata.version(__package__)

from .documentor import (
    catalogue as create_catalogue,
)
from .documentor import (
    table as create_table,
)
from .labeller import label as label
from .loader import load as load
from .validator import validate as validate



from rdflib.namespace import DefinedNamespace, Namespace
from rdflib.term import URIRef


class MRR(DefinedNamespace):
    _NS = Namespace("https://prez.dev/ManifestResourceRoles/")
    _fail = True

    CatalogueData: URIRef
    CatalogueModel: URIRef
    ResourceData: URIRef
    ResourceModel: URIRef
    CatalogueAndResourceModel: URIRef
    CompleteCatalogueAndResourceLabels: URIRef
    IncompleteCatalogueAndResourceLabels: URIRef


class OLIS(DefinedNamespace):
    _NS = Namespace("https://olis.dev/")
    _fail = True

    NamedGraph: URIRef
    RealGraph: URIRef
    SystemGraph: URIRef
    VirtualGraph: URIRef

    isAliasFor: URIRef


class PREZ(DefinedNamespace):
    _NS = Namespace("https://prez.dev/")
    _fail = True

    Manifest: URIRef



from pathlib import Path

import pytest
from kurra.db import sparql, upload
from rdflib import Graph
from rdflib.compare import isomorphic
from typer.testing import CliRunner

from prezmanifest.cli import app
from prezmanifest.labeller import LabellerOutputTypes, label
from tests.fuseki.conftest import fuseki_container

runner = CliRunner()


def test_label_iris():
    with pytest.raises(ValueError):
        label(
            Path(__file__).parent / "demo-vocabs/manifest-labels-none.ttl",
            output_type="x",
        )

    iris = label(
        Path(__file__).parent / "demo-vocabs/manifest-labels-none.ttl",
        output_type=LabellerOutputTypes.iris,
    )

    assert len(iris) == 31

    iris = label(
        Path(__file__).parent / "demo-vocabs/manifest-labels-some.ttl",
        output_type=LabellerOutputTypes.iris,
    )

    assert len(iris) == 16

    iris = label(
        Path(__file__).parent / "demo-vocabs/manifest-labels-all.ttl",
        output_type=LabellerOutputTypes.iris,
    )

    assert len(iris) == 0


def test_label_iris_sparql(fuseki_container):
    SPARQL_ENDPOINT = f"http://localhost:{fuseki_container.get_exposed_port(3030)}/ds"
    upload(
        SPARQL_ENDPOINT,
        Path(__file__).parent / "demo-vocabs/labels-some.ttl",
        graph_id="http://test",
    )

    iris = label(
        Path(__file__).parent / "demo-vocabs/manifest-labels-none.ttl",
        output_type=LabellerOutputTypes.iris,
        additional_context=SPARQL_ENDPOINT,
    )

    assert len(iris) == 15


def test_label_rdf_file():
    with pytest.raises(ValueError):
        label(
            Path(__file__).parent / "demo-vocabs/manifest-labels-none.ttl",
            output_type=LabellerOutputTypes.rdf,
        )

    rdf = label(
        Path(__file__).parent / "demo-vocabs/manifest-labels-none.ttl",
        output_type=LabellerOutputTypes.rdf,
        additional_context=Path(__file__).parent / "demo-vocabs/labels-some.ttl",
    )

    assert len(rdf) == 37


def test_label_rdf_sparql(fuseki_container):
    SPARQL_ENDPOINT = f"http://localhost:{fuseki_container.get_exposed_port(3030)}/ds"

    sparql(SPARQL_ENDPOINT, "DROP SILENT GRAPH <http://test>")

    rdf = label(
        Path(__file__).parent / "demo-vocabs/manifest.ttl",
        LabellerOutputTypes.rdf,
        SPARQL_ENDPOINT,
    )

    assert len(rdf) == 0

    rdf = label(
        Path(__file__).parent / "demo-vocabs/manifest-labels-none.ttl",
        LabellerOutputTypes.rdf,
        SPARQL_ENDPOINT,
    )

    assert len(rdf) == 0

    upload(
        SPARQL_ENDPOINT,
        Path(__file__).parent / "demo-vocabs/labels-some.ttl",
        graph_id="http://test",
    )

    rdf = label(
        Path(__file__).parent / "demo-vocabs/manifest-labels-none.ttl",
        LabellerOutputTypes.rdf,
        SPARQL_ENDPOINT,
    )

    assert len(rdf) == 37

    sparql(SPARQL_ENDPOINT, "DROP GRAPH <http://test>")

    upload(
        SPARQL_ENDPOINT,
        Path(__file__).parent / "demo-vocabs/labels-all.ttl",
        graph_id="http://test",
    )

    rdf = label(
        Path(__file__).parent / "demo-vocabs/manifest-labels-none.ttl",
        LabellerOutputTypes.rdf,
        SPARQL_ENDPOINT,
    )

    assert len(rdf) == 63


def test_label_manifest(fuseki_container):
    SPARQL_ENDPOINT = f"http://localhost:{fuseki_container.get_exposed_port(3030)}/ds"

    sparql(SPARQL_ENDPOINT, "DROP SILENT GRAPH <http://test>")

    upload(
        SPARQL_ENDPOINT,
        Path(__file__).parent / "demo-vocabs/_background/labels.ttl",
        graph_id="http://test",
    )

    original_manifest_path = (
        Path(__file__).parent / "demo-vocabs/manifest-labels-none.ttl"
    )
    original_manifest_contents = original_manifest_path.read_text()

    label(
        Path(__file__).parent / "demo-vocabs/manifest-labels-none.ttl",
        # output="manifest" is default
        additional_context=SPARQL_ENDPOINT,
    )

    expected_updated_manifest = Graph().parse(
        data="""
        PREFIX mrr: <https://prez.dev/ManifestResourceRoles/>
        PREFIX prez: <https://prez.dev/>
        PREFIX prof: <http://www.w3.org/ns/dx/prof/>
        PREFIX schema: <https://schema.org/>
        
        []    a prez:Manifest ;
            prof:hasResource
                [
                    prof:hasArtifact "labels-additional.ttl" ;
                    prof:hasRole mrr:IncompleteCatalogueAndResourceLabels ;
                ] ,
                [
                    prof:hasArtifact "catalogue.ttl" ;
                    prof:hasRole mrr:CatalogueData ;
                    schema:description "The definition of, and medata for, the container which here is a dcat:Catalog object" ;
                    schema:name "Catalogue Definition" ;
                ] ,
                [
                    prof:hasArtifact "vocabs/*.ttl" ;
                    prof:hasRole mrr:ResourceData ;
                    schema:description "skos:ConceptScheme objects in RDF (Turtle) files in the vocabs/ folder" ;
                    schema:name "Resource Data" ;
                ] ;
        .
        """
    )

    resulting_manifest = Graph().parse(original_manifest_path)

    assert isomorphic(resulting_manifest, expected_updated_manifest)

    # replace this test's results and with original
    Path(original_manifest_path.parent / "labels-additional.ttl").unlink()
    original_manifest_path.unlink()
    original_manifest_path.write_text(original_manifest_contents)


def test_label_iris_mainEntity():
    iris = label(
        Path(__file__).parent / "demo-vocabs/manifest-mainEntity.ttl",
        output_type=LabellerOutputTypes.iris,
    )

    assert len(iris) == 14


def test_label_cli_iris():
    result = runner.invoke(
        app,
        [
            "label",
            "iris",
            str(Path(__file__).parent / "demo-vocabs/manifest-labels-none.ttl"),
        ],
    )
    assert len(result.stdout.splitlines()) == 31


def test_label_cli_rdf(fuseki_container):
    SPARQL_ENDPOINT = f"http://localhost:{fuseki_container.get_exposed_port(3030)}/ds"

    sparql(SPARQL_ENDPOINT, "DROP GRAPH <http://test>")

    upload(
        SPARQL_ENDPOINT,
        Path(__file__).parent / "demo-vocabs/_background/labels.ttl",
        graph_id="http://test",
    )

    upload(
        SPARQL_ENDPOINT,
        Path(__file__).parent / "demo-vocabs/labels-2.ttl",
        graph_id="http://test",
        append=True,
    )

    result = runner.invoke(
        app,
        [
            "label",
            "rdf",
            str(Path(__file__).parent / "demo-vocabs/manifest-labels-none.ttl"),
            SPARQL_ENDPOINT,
        ],
    )
    assert len(Graph().parse(data=result.stdout, format="turtle")) == 52


# TODO: fix not working test
# def test_label_cli_manifest(fuseki_container):
#     SPARQL_ENDPOINT = f"http://localhost:{fuseki_container.get_exposed_port(3030)}/ds"
#
#     sparql(SPARQL_ENDPOINT, "DROP SILENT GRAPH <http://test>")
#
#     upload(
#         SPARQL_ENDPOINT,
#         Path(__file__).parent / "demo-vocabs/_background/labels.ttl",
#         graph_id="http://test",
#     )
#
#     original_manifest_path = (
#             Path(__file__).parent / "demo-vocabs/manifest-labels-none.ttl"
#     )
#     original_manifest_contents = original_manifest_path.read_text()
#
#     runner.invoke(
#         app,
#         [
#             "label",
#             "manifest",
#             str(original_manifest_path),
#             SPARQL_ENDPOINT
#         ],
#     )
#
#     expected_updated_manifest = Graph().parse(
#         data="""
#         PREFIX mrr: <https://prez.dev/ManifestResourceRoles/>
#         PREFIX prez: <https://prez.dev/>
#         PREFIX prof: <http://www.w3.org/ns/dx/prof/>
#         PREFIX schema: <https://schema.org/>
#
#         []    a prez:Manifest ;
#             prof:hasResource
#                 [
#                     prof:hasArtifact "labels-additional.ttl" ;
#                     prof:hasRole mrr:IncompleteCatalogueAndResourceLabels ;
#                 ] ,
#                 [
#                     prof:hasArtifact "catalogue.ttl" ;
#                     prof:hasRole mrr:CatalogueData ;
#                     schema:description "The definition of, and medata for, the container which here is a dcat:Catalog object" ;
#                     schema:name "Catalogue Definition" ;
#                 ] ,
#                 [
#                     prof:hasArtifact "vocabs/*.ttl" ;
#                     prof:hasRole mrr:ResourceData ;
#                     schema:description "skos:ConceptScheme objects in RDF (Turtle) files in the vocabs/ folder" ;
#                     schema:name "Resource Data" ;
#                 ] ,
#                 [
#                     prof:hasArtifact "https://raw.githubusercontent.com/RDFLib/prez/refs/heads/main/prez/reference_data/profiles/ogc_records_profile.ttl" ;
#                     prof:hasRole mrr:CatalogueAndResourceModel ;
#                     schema:description "The default Prez profile for Records API" ;
#                     schema:name "Profile Definition" ;
#                 ] ;
#         .
#         """
#     )
#
#     resulting_manifest = Graph().parse(original_manifest_path)
#
#     assert isomorphic(resulting_manifest, expected_updated_manifest)
#
#     # replace this test's results and with original
#     Path(original_manifest_path.parent / "labels-additional.ttl").unlink()
#     original_manifest_path.unlink()
#     original_manifest_path.write_text(original_manifest_contents)



import json
import shutil
from pathlib import Path

from kurra.db import sparql
from typer.testing import CliRunner

from prezmanifest.loader import load
from prezmanifest.syncer import sync
from prezmanifest.utils import artifact_file_name_from_graph_id
from tests.fuseki.conftest import fuseki_container

runner = CliRunner(mix_stderr=False)
from prezmanifest.cli import app


def test_sync(fuseki_container):
    SPARQL_ENDPOINT = f"http://localhost:{fuseki_container.get_exposed_port(3030)}/ds"
    MANIFEST_FILE_LOCAL = Path(__file__).parent / "local/manifest.ttl"
    MANIFEST_FILE_REMOTE = Path(__file__).parent / "remote/manifest.ttl"
    MANIFEST_ROOT = Path(__file__).parent / "local"

    # make copies of files that will be overwritten
    shutil.copy(MANIFEST_FILE_LOCAL, MANIFEST_FILE_LOCAL.with_suffix(".ttx"))
    shutil.copy(MANIFEST_ROOT / "catalogue.ttl", MANIFEST_ROOT / "catalogue.ttx")
    shutil.copy(MANIFEST_ROOT / "artifact6.ttl", MANIFEST_ROOT / "artifact6.ttx")

    # ensure the SPARQL store's clear
    sparql(SPARQL_ENDPOINT, "DROP ALL")

    # load it with remote data
    load(MANIFEST_FILE_REMOTE, SPARQL_ENDPOINT)

    a = sync(
        MANIFEST_FILE_LOCAL,
        SPARQL_ENDPOINT,
    )

    # check status before sync
    assert a[str(MANIFEST_ROOT / "artifacts/artifact1.ttl")]["direction"] == "same"
    assert a[str(MANIFEST_ROOT / "artifacts/artifact2.ttl")]["direction"] == "upload"
    assert a[str(MANIFEST_ROOT / "artifacts/artifact3.ttl")]["direction"] == "upload"
    assert a[str(MANIFEST_ROOT / "artifact4.ttl")]["direction"] == "upload"
    assert a[str(MANIFEST_ROOT / "artifact5.ttl")]["direction"] == "add-remotely"
    assert a[str(MANIFEST_ROOT / "artifact6.ttl")]["direction"] == "download"
    assert a[str(MANIFEST_ROOT / "artifact7.ttl")]["direction"] == "upload"
    assert a["http://example.com/dataset/8"]["direction"] == "add-locally"
    assert a[str(MANIFEST_ROOT / "catalogue.ttl")]["direction"] == "same"

    # run sync again, performing no actions to just get updated status
    a = sync(MANIFEST_FILE_LOCAL, SPARQL_ENDPOINT, None, False, False, False, False)

    # check status after sync
    assert a[str(MANIFEST_ROOT / "artifacts/artifact1.ttl")]["direction"] == "same"
    assert a[str(MANIFEST_ROOT / "artifacts/artifact2.ttl")]["direction"] == "same"
    assert a[str(MANIFEST_ROOT / "artifacts/artifact3.ttl")]["direction"] == "same"
    assert a[str(MANIFEST_ROOT / "artifact4.ttl")]["direction"] == "same"
    assert a[str(MANIFEST_ROOT / "artifact5.ttl")]["direction"] == "same"
    assert a[str(MANIFEST_ROOT / "artifact6.ttl")]["direction"] == "same"
    assert a[str(MANIFEST_ROOT / "artifact7.ttl")]["direction"] == "same"
    assert (
        a[
            str(
                MANIFEST_ROOT
                / artifact_file_name_from_graph_id("http://example.com/dataset/8")
            )
        ]["direction"]
        == "same"
    )
    assert a[str(MANIFEST_ROOT / "catalogue.ttl")]["direction"] == "same"

    # tidy up
    shutil.move(MANIFEST_ROOT / "manifest.ttx", MANIFEST_FILE_LOCAL)
    shutil.move(MANIFEST_ROOT / "catalogue.ttx", MANIFEST_ROOT / "catalogue.ttl")
    shutil.move(MANIFEST_ROOT / "artifact6.ttx", MANIFEST_ROOT / "artifact6.ttl")
    Path(MANIFEST_ROOT / "http--example.com-dataset-8.ttl").unlink()


def test_sync_cli(fuseki_container):
    MANIFEST_FILE_REMOTE = Path(__file__).parent / "remote/manifest.ttl"
    SPARQL_ENDPOINT = f"http://localhost:{fuseki_container.get_exposed_port(3030)}/ds"

    # ensure the SPARQL store's clear
    sparql(SPARQL_ENDPOINT, "DROP ALL")

    raw_output = str(
        runner.invoke(
            app, ["sync", str(MANIFEST_FILE_REMOTE), SPARQL_ENDPOINT, "-f", "json"]
        ).stdout
    )

    r = json.loads(raw_output)

    assert str(MANIFEST_FILE_REMOTE.parent / "catalogue.ttl") in r.keys()

    # test cli pretty formatting
    raw_output = str(
        runner.invoke(app, ["sync", str(MANIFEST_FILE_REMOTE), SPARQL_ENDPOINT]).stdout
    )
    assert "Main Entity" in raw_output



from prezmanifest.cli.app import app
from prezmanifest.cli.commands.label import app as label_app
from prezmanifest.cli.commands.validate import validate_command

app.add_typer(label_app, name="label")

from prezmanifest.cli.commands.document import app as document_app

app.add_typer(document_app, name="document")

from prezmanifest.cli.commands.load import app as load_app

app.add_typer(load_app, name="load")

from prezmanifest.cli.commands.sync import sync_command



"""Validate a Prez Manifest file.

This script performs both SHACL validation to ensure the Manifest is valid according to the Prez Manifest
specification (see https://prez.dev/manifest/) and checks that all the resources indicated by the Manifest
- whether local files/folders or remote resources on the Internet - are reachable.

~$ python validate.py {MANIFEST_FILE_PATH}"""

from pathlib import Path

import httpx
from kurra.utils import load_graph
from pyparsing import Literal
from pyshacl import validate as shacl_validate
from rdflib import BNode, Dataset, Graph
from rdflib.namespace import DCTERMS, PROF, SDO

from prezmanifest.definednamespaces import MRR
from prezmanifest.utils import get_files_from_artifact, get_validator_graph


class ManifestValidationError(Exception):
    pass


def validate(manifest: Path) -> Graph:
    # can't use get_manifest_paths_and_graph() here as that function uses validate()
    manifest_path = manifest
    manifest_root = Path(manifest).parent.resolve()
    manifest_graph = load_graph(manifest)

    def literal_resolves_as_file_folder_or_url(lit: Literal):
        l_str = str(lit)
        if l_str.startswith("http") and "://" in l_str:
            r = httpx.get(l_str)
            if 200 <= r.status_code < 400:
                pass
            else:
                raise ManifestValidationError(
                    f"Remote content link non-resolving: {l_str}"
                )
        elif "*" in l_str:
            glob_parts = l_str.split("*")
            dir = Path(manifest_root / Path(glob_parts[0]))
            if not Path(dir).is_dir():
                raise ManifestValidationError(
                    f"The content link {l_str} is not a directory"
                )
        else:
            # It must be a local
            if not (manifest_root / l_str).is_file():
                raise ManifestValidationError(
                    f"Content link {manifest_root / l_str} is invalid - not a file"
                )

    def shacl_validate_resource(data_graph, shacl_graph) -> (bool, str | None):
        valid, v_graph, v_text = shacl_validate(
            data_graph, shacl_graph=shacl_graph, allow_warnings=True
        )
        if valid:
            return True, None
        else:
            return False, v_text

    ME = Path(__file__)

    # SHACL validation

    mrr_vocab_graph = load_graph(ME.parent / "mrr.ttl")
    shacl_graph = load_graph(ME.parent / "validator.ttl")
    valid, error_msg = shacl_validate_resource(
        manifest_graph + mrr_vocab_graph, shacl_graph
    )
    if not valid:
        raise ManifestValidationError(f"Manifest Shapes invalid:\n\n{error_msg}")

    # get labels graph for SHACL validation
    context_graph = Graph()
    for s, o in manifest_graph.subject_objects(PROF.hasResource):
        for role in manifest_graph.objects(o, PROF.hasRole):
            # The data files & background - must be processed after Catalogue
            if role in [
                MRR.CompleteCatalogueAndResourceLabels,
                MRR.IncompleteCatalogueAndResourceLabels,
            ]:
                for artifact in manifest_graph.objects(o, PROF.hasArtifact):
                    for f in get_files_from_artifact(
                        (manifest_path, manifest_root, manifest_graph), artifact
                    ):
                        if not f.is_file():
                            raise ManifestValidationError(f"Artifact {f} is not a file")

                        if str(f.name).endswith(".ttl"):
                            context_graph += load_graph(f)
                        elif str(f.name).endswith(".trig"):  # TODO: test this option
                            d = Dataset()
                            d.parse(f, format="trig")
                            for g in d.graphs:
                                context_graph += g

    # Content link validation
    for s, o in manifest_graph.subject_objects(PROF.hasResource):
        # see if there's a conformance claim for the resource
        cc = manifest_graph.value(subject=o, predicate=DCTERMS.conformsTo)

        for artifact in manifest_graph.objects(o, PROF.hasArtifact):
            if isinstance(artifact, BNode):
                content_location = manifest_graph.value(
                    subject=artifact, predicate=SDO.contentLocation
                )
                # main_entity = manifest_graph.value(subject=artifact, predicate=SDO.mainEntity)
            else:
                content_location = artifact

            # ensure the artifact resolves
            literal_resolves_as_file_folder_or_url(content_location)

            # if we now have a CC for the resource or the artifact, use it
            for file in get_files_from_artifact(
                (manifest_path, manifest_root, manifest_graph), content_location
            ):
                # if there is a conformance claim for this Resource, use it, if not, check if the artifact has one
                if cc is None:
                    cc = manifest_graph.value(
                        subject=artifact, predicate=DCTERMS.conformsTo
                    )
                    artifact_cc = True
                else:
                    artifact_cc = False

                if cc is not None:
                    data_graph = load_graph(manifest_root / file)
                    if context_graph is not None:
                        data_graph += context_graph
                    valid, error_msg = shacl_validate_resource(
                        data_graph,
                        get_validator_graph(
                            (manifest_path, manifest_root, manifest_graph), cc
                        ),
                    )
                    if not valid:
                        raise ManifestValidationError(
                            f"Resource {content_location} Shapes invalid according to conformance claim:\n\n{error_msg}"
                        )

                if artifact_cc:
                    cc = None

    return manifest_graph



from collections.abc import Generator
from enum import Enum
from pathlib import Path

import httpx
from dateutil.parser import parse as date_parse
from kurra.db import sparql
from kurra.file import load_graph
from kurra.sparql import query
from rdflib import BNode, Dataset, Graph, Literal, Node, URIRef
from rdflib.namespace import DCAT, OWL, PROF, RDF, SDO, SKOS

import prezmanifest
from prezmanifest.definednamespaces import MRR, PREZ

KNOWN_PROFILES = {
    URIRef("http://www.opengis.net/def/geosparql"): {
        "path": Path(__file__).parent / "validator-geosparql-1.1.ttl",
        "main_entity_classes": [SDO.Dataset, DCAT.Dataset],
    },
    URIRef("https://data.idnau.org/pid/cp"): {
        "path": Path(__file__).parent / "validator-idn-cp.ttl",
        "main_entity_classes": [SDO.Dataset, DCAT.Dataset],
    },
    URIRef("https://w3id.org/profile/vocpub"): {
        "path": Path(__file__).parent / "validator-vocpub-4.10.ttl",
        "main_entity_classes": [SKOS.ConceptScheme],
    },
}

KNOWN_ENTITY_CLASSES = [
    SKOS.ConceptScheme,
    OWL.Ontology,
    DCAT.Resource,
    SDO.CreativeWork,
    SDO.Dataset,
    DCAT.Dataset,
    SDO.DefinedTerm,
    SDO.DataCatalog,
    DCAT.Catalog,
]

ENTITY_CLASSES_PER_PROFILE = {
    "": "",
}


def path_or_url(s: str) -> Path | str:
    """Converts a string into a Path, preserving http(s)://..."""
    if s.startswith("http") and "://" in str(s):
        return s
    else:
        return Path(s)


def localise_path(p: Path | str, root: Path) -> Path:
    if str(p).startswith("http") and "://" in str(p):
        return p
    else:
        return Path(str(p).replace(str(root) + "/", ""))


def absolutise_path(p: Path | str, root: Path) -> Path | str:
    if str(p).startswith("http") and "://" in str(p):
        if "://" not in str(p):
            return str(p).replace(":/", "://")
        else:
            return str(p)
    else:
        return root / p


def get_files_from_artifact(
    manifest: Path | tuple[Path, Path, Graph], artifact: Node
) -> list[Path | str] | Generator[Path]:
    """Returns an iterable (list or generator) of Path objects for files within an artifact literal.

    This function will correctly interpret artifacts such as 'file.ttl', '*.ttl', '**/*.trig' etc.
    """
    manifest_path, manifest_root, manifest_graph = get_manifest_paths_and_graph(
        manifest
    )

    if str(artifact).startswith("http") and "://" in str(artifact):
        return [str(artifact)]
    elif isinstance(artifact, Literal):
        if "*" not in str(artifact):
            return [manifest_root / path_or_url(str(artifact))]
        else:
            artifact_str = str(artifact)
            glob_marker_location = artifact_str.find("*")
            glob_parts = [
                artifact_str[:glob_marker_location],
                artifact_str[glob_marker_location:],
            ]
            return Path(manifest_root / path_or_url(glob_parts[0])).rglob(glob_parts[1])
    elif isinstance(artifact, BNode):
        contentLocation = manifest_graph.value(
            subject=artifact, predicate=SDO.contentLocation
        )
        if str(contentLocation).startswith("http") and "://" in str(contentLocation):
            return [str(contentLocation)]
        else:
            return [manifest_root / str(contentLocation)]
    else:
        raise TypeError(f"Unsupported artifact type: {type(artifact)}")


def get_identifier_from_file(file: Path) -> list[URIRef]:
    """Returns a list if RDFLib graph identifier (URIRefs) from a triples or quads file
    for all owl:Ontology and skos:ConceptScheme objects"""
    if file.name.endswith(".ttl"):
        g = Graph().parse(file)
        for entity_class in KNOWN_ENTITY_CLASSES:
            v = g.value(predicate=RDF.type, object=entity_class)
            if v is not None:
                return [v]
    elif file.name.endswith(".trig"):
        gs = []
        d = Dataset()
        d.parse(file, format="trig")
        for g in d.graphs():
            gs.append(g.identifier)
        return gs
    else:
        return []


def get_validator_graph(
    manifest: Path | tuple[Path, Path, Graph], iri_or_path: URIRef | Literal
) -> Graph:
    """Returns the graph of a validator from either the path of a SHACL file or a known IRI->profile validator file"""
    manifest_path, manifest_root, manifest_graph = get_manifest_paths_and_graph(
        manifest
    )

    if isinstance(iri_or_path, URIRef):
        if iri_or_path not in KNOWN_PROFILES.keys():
            raise ValueError(
                f"You have specified conformance to an unknown profile. Known profiles are {', '.join(KNOWN_PROFILES.keys())}"
            )
        return load_graph(KNOWN_PROFILES[iri_or_path]["path"])

    else:
        return load_graph(absolutise_path(iri_or_path, manifest_root))


def get_manifest_paths_and_graph(
    manifest: Path | tuple[Path, Path, Graph],
) -> (Path, Graph):
    """Reads either a Manifest file from a Path, or a Manifest file from a Path and its root directory,
    a Path, and the Manifest as a deserialized Graph and returns the Manifest Path, its root dir as a Path
    and its content as a Graph"""
    if isinstance(manifest, Path):
        manifest_path = manifest
        manifest_root = Path(manifest).parent.resolve()
        manifest_graph = prezmanifest.validate(manifest)
    else:  # (Path, Path, Graph)
        manifest_path = manifest[0]
        manifest_root = manifest[1]
        manifest_graph = manifest[2]

    return manifest_path, manifest_root, manifest_graph


def get_catalogue_iri_from_manifest(
    manifest: Path | tuple[Path, Path, Graph],
) -> URIRef:
    manifest_path, manifest_root, manifest_graph = get_manifest_paths_and_graph(
        manifest
    )

    for m in manifest_graph.subjects(RDF.type, PREZ.Manifest):
        for r in manifest_graph.objects(m, PROF.hasResource):
            for role in manifest_graph.objects(r, PROF.hasRole):
                if role == MRR.CatalogueData:
                    artifacts = manifest_graph.objects(r, PROF.hasArtifact)
                    for x in artifacts:
                        a = x
                    if isinstance(a, Literal):
                        a_graph = load_graph(manifest_root / str(a))
                        return a_graph.value(
                            predicate=RDF.type, object=DCAT.Catalog
                        ) or a_graph.value(predicate=RDF.type, object=SDO.DataCatalog)

    raise ValueError(f"No catalogue object IRI found in Manifest {manifest_root}")


def target_contains_this_manifests_catalogue(
    manifest: Path | tuple[Path, Path, Graph],
    sparql_endpoint: str = None,
    http_client: httpx.Client | None = None,
) -> bool:
    # get the IRI of the catalogue from the manifest
    cat_iri = get_catalogue_iri_from_manifest(manifest)

    q = """
        PREFIX olis: <https://olis.dev/>
        
        ASK
        WHERE {
          GRAPH olis:SystemGraph {
            VALUES ?graph_type {
              olis:RealGraph 
              olis:VirtualGraph
            }
            <xxx> a ?graph_type
          }
        }
        """.replace(
        "xxx", cat_iri
    )

    return query(
        sparql_endpoint, q, http_client, return_python=True, return_bindings_only=True
    )


def make_httpx_client(
    sparql_username: str = None,
    sparql_password: str = None,
):
    auth = None
    if sparql_username:
        if sparql_password:
            auth = httpx.BasicAuth(sparql_username, sparql_password)
    return httpx.Client(auth=auth)


def get_main_entity_iri_of_artifact(
    artifact: Path,
    manifest: Path | tuple[Path, Path, Graph],
    artifact_graph: Graph = None,
    cc: URIRef = None,
    atype: URIRef = None,
) -> URIRef:
    manifest_path, manifest_root, manifest_graph = get_manifest_paths_and_graph(
        manifest
    )
    artifact_path = absolutise_path(artifact, manifest_root)
    known_entity_classes = []
    if cc is not None:
        if cc in KNOWN_PROFILES.keys():
            for m_e_c in KNOWN_PROFILES[cc]["main_entity_classes"]:
                known_entity_classes.append(str(m_e_c))

    if atype is not None:
        known_entity_classes.append(str(atype))

    if len(known_entity_classes) < 1:
        known_entity_classes = [str(x) for x in KNOWN_ENTITY_CLASSES]

    known_entity_classes_str = f"<{'>\n                <'.join(known_entity_classes)}>"
    q = f"""
        SELECT ?me
        WHERE {{
            VALUES ?t {{
                {known_entity_classes_str.strip()}
            }}
            ?me a ?t .
        }}
        """
    mes = []

    g = artifact_graph if artifact_graph is not None else load_graph(artifact_path)

    if not isinstance(g, Graph):
        raise ValueError(f"Could not load a graph of the artifact at {artifact_path}")

    for r in query(g, q, return_python=True, return_bindings_only=True):
        if r.get("me"):
            mes.append(r["me"]["value"])

    if len(mes) != 1:
        if len(mes) > 1:
            raise ValueError(
                f"The artifact at {artifact_path} has more than one Main Entity: {', '.join(mes)} "
                f"based on the class {cc if cc is not None else '(none given)'}. There must only be one."
            )
        else:
            raise ValueError(
                f"The artifact at {artifact_path} has no recognizable Main Entity, "
                f"based on the classes {cc if cc is not None else '(none given)'}. There must be one."
            )

    return URIRef(mes[0])


def get_version_indicators_local(
    manifest: Path | tuple[Path, Path, Graph], artifact: Path, version_indicators: dict
):
    manifest_path, manifest_root, manifest_graph = get_manifest_paths_and_graph(
        manifest
    )
    artifact_path = absolutise_path(artifact, manifest_root)
    artifact_graph = load_graph(artifact_path)

    # if we aren't given a Main Entity, let's look for one using the Main Entity Classes
    if version_indicators.get("main_entity") is None:
        version_indicators["main_entity"] = get_main_entity_iri_of_artifact(
            artifact,
            (manifest_path, manifest_root, manifest_graph),
            artifact_graph,
            version_indicators.get("conformance_claim"),
            version_indicators.get("additional_type"),
        )

    # if we have a Main Entity at this point, we can get the content-based Indicators
    if version_indicators.get("main_entity") is not None:
        q = f"""
            PREFIX dcterms: <http://purl.org/dc/terms/>
            PREFIX owl: <http://www.w3.org/2002/07/owl#>
            PREFIX schema: <https://schema.org/>
            
            SELECT ?md ?vi ?v
            WHERE {{
                OPTIONAL {{
                    <{version_indicators["main_entity"]}> dcterms:modified|schema:dateModified ?md .
                }}
                
                OPTIONAL {{
                    <{version_indicators["main_entity"]}> owl:versionIRI ?vi .
                }}
                
                OPTIONAL {{
                 <{version_indicators["main_entity"]}> owl:versionInfo|schema:version|dcterms:hasVersion ?v .
             }}
            }}
            """
        # only use values for Version Indicators if not already present - i.e. from the manifest
        for r in query(
            artifact_graph, q, return_python=True, return_bindings_only=True
        ):
            if version_indicators.get("modified_date") is None:
                if r.get("md") is not None:
                    version_indicators["modified_date"] = date_parse(r["md"]["value"])
            if version_indicators.get("version_iri") is None:
                if r.get("vi") is not None:
                    version_indicators["version_iri"] = r["vi"]["value"]
            if version_indicators.get("version_info") is None:
                if r.get("v") is not None:
                    version_indicators["version_info"] = r["v"]["value"]
    # if not, we may still get file-based indicators
    if artifact_path.is_file():
        version_indicators["file_size"] = artifact_path.stat().st_size

    return


def get_version_indicators_sparql(
    main_entity: str,
    sparql_endpoint: str,
    http_client: httpx.Client | None = None,
) -> dict:
    if not sparql_endpoint.startswith("http") and "://" in str(sparql_endpoint):
        raise ValueError(
            f"The sparql_endpoint you have supplied does not look valid: {sparql_endpoint}"
        )

    indicators = {
        "modified_date": None,
        "version_info": None,
        "version_iri": None,
        "file_size": None,
        "main_entity": main_entity,
    }

    q = f"""
        PREFIX dcterms: <http://purl.org/dc/terms/>
        PREFIX owl: <http://www.w3.org/2002/07/owl#>
        PREFIX schema: <https://schema.org/>

        SELECT ?md ?vi ?v
        WHERE {{
            GRAPH ?g {{
                VALUES ?me {{
                    <{main_entity}>
                    <{main_entity + "-catalogue"}>
                }}
                OPTIONAL {{
                    ?me dcterms:modified|schema:dateModified ?md .
                }}
    
                OPTIONAL {{
                    ?me owl:versionIRI ?vi .
                }}
    
                OPTIONAL {{
                    ?me owl:versionInfo|schema:version|dcterms:hasVersion ?v .
                }}
            }}
        }}
        """
    res = sparql(
        sparql_endpoint, q, http_client, return_python=True, return_bindings_only=True
    )
    if len(res) == 0:
        raise ValueError(
            "The system got no results from its querying of the SPARQL endpoint"
        )

    for r in res:
        if r.get("md") is not None:
            indicators["modified_date"] = date_parse(r["md"]["value"])
        if r.get("vi") is not None:
            indicators["version_iri"] = r["vi"]["value"]
        if r.get("v") is not None:
            indicators["version_info"] = r["v"]["value"]

    return indicators


class VersionIndicatorComparison(Enum):
    First = "first"
    Second = "second"
    Neither = "neither"
    CantCalculate = "cant_calculate"


def compare_version_indicators(first: dict, second: dict) -> VersionIndicatorComparison:
    """Compares Modified Date, Version IRI & Version info for each and returns latest"""

    """Even weighted aggregate score for each version indicator"""
    first_score = 0
    second_score = 0
    has_modified_date_comparison = first.get("modified_date") and second.get(
        "modified_date"
    )
    has_version_iri_comparison = first.get("version_iri") and second.get("version_iri")
    has_version_info_comparison = first.get("version_info") and second.get(
        "version_info"
    )

    if (
        not has_modified_date_comparison
        and not has_version_iri_comparison
        and not has_version_info_comparison
    ):
        return VersionIndicatorComparison.CantCalculate

    if has_modified_date_comparison:
        if first["modified_date"] > second["modified_date"]:
            first_score += 1
        elif first["modified_date"] == second["modified_date"]:
            pass
        else:
            second_score += 1

    if has_version_iri_comparison:
        if first["version_iri"] > second["version_iri"]:
            first_score += 1
        elif first["version_iri"] == second["version_iri"]:
            pass
        else:
            second_score += 1

    if has_version_info_comparison:
        if first["version_info"] > second["version_info"]:
            first_score += 1
        elif first["version_info"] == second["version_info"]:
            pass
        else:
            second_score += 1

    # TODO: add test for file_size, Git version etc.

    if first_score > second_score:
        return VersionIndicatorComparison.First
    elif second_score == first_score:
        return VersionIndicatorComparison.Neither
    else:
        return VersionIndicatorComparison.Second


def which_is_more_recent(
    version_indicators: dict,
    sparql_endpoint: str = None,
    http_client: httpx.Client | None = None,
) -> VersionIndicatorComparison:
    """Tests to see if the given artifact is more recent than a previously stored copy of its content"""

    remote = get_version_indicators_sparql(
        version_indicators["main_entity"], sparql_endpoint, http_client
    )

    return compare_version_indicators(version_indicators, remote)


def denormalise_artifacts(manifest: Path | tuple[Path, Path, Graph] = None) -> dict:
    """For each Artifiact in the Manifest, return a dict:

    Artifact path,
    Main Entity,
    Conformance Claims
    Date Modified
    Version IRI
    Version Info
    Role
    Version Indicators

    of each asset's path, main entity IRI, manifest role and version indicators"""

    artifacts_info = {}

    manifest_path, manifest_root, manifest_graph = get_manifest_paths_and_graph(
        manifest
    )

    # for each artifact, get what we can directly from the Manifest
    q = """
        PREFIX dcterms: <http://purl.org/dc/terms/>
        PREFIX mrr: <https://prez.dev/ManifestResourceRoles/>
        PREFIX owl: <http://www.w3.org/2002/07/owl#>
        PREFIX prez: <https://prez.dev/>
        PREFIX prof: <http://www.w3.org/ns/dx/prof/>
        PREFIX schema: <https://schema.org/>

        SELECT ?a ?me ?cc ?atype ?dm ?vi ?v ?r
        WHERE {
            # if the Resource has a Blank Node artifact, it must provide the Main Entity IRI
            {
                ?x 
                    prof:hasArtifact ?bn ;
                    prof:hasRole ?r ;
                .
                
                ?bn 
                    schema:mainEntity ?me ;
                    schema:contentLocation ?a ;
                .
                
                OPTIONAL {
                    ?bn dcterms:conformsTo ?cc_local .
                }
                
                OPTIONAL {
                    ?x dcterms:conformsTo ?cc_resource .
                }
                
                OPTIONAL {
                    ?bn schema:additionalType ?atype_local .
                }
                
                OPTIONAL {
                    ?x schema:additionalType ?atype_resource .
                }                
                
                OPTIONAL {
                    ?bn schema:dateModified ?dm .
                } 
                
                OPTIONAL {
                    ?bn owl:versionIRI ?vi .
                }
                
                OPTIONAL {
                    ?bn owl:versionInfo|schema:version|dcterms:hasVersion ?v .
                }     
                
                BIND(COALESCE(?cc_local, ?cc_resource) AS ?cc)
                
                BIND(COALESCE(?atype_local, ?atype_resource) AS ?atype)
                   
                FILTER isBLANK(?bn)
            }
            UNION 
            {
                ?x 
                    prof:hasArtifact ?a ;
                    prof:hasRole ?r ;
                .
                
                OPTIONAL {
                    ?x dcterms:conformsTo ?cc .
                }
                
                OPTIONAL {
                    ?x schema:additionalType ?atype .
                }                
                
                FILTER isLITERAL(?a)
            }
        }
        """

    for r in query(manifest_graph, q, return_python=True, return_bindings_only=True):
        artifact = path_or_url(r["a"]["value"])
        files = get_files_from_artifact(
            (manifest_path, manifest_root, manifest_graph), Literal(artifact)
        )

        for file in files:
            me = URIRef(r["me"]["value"]) if r.get("me") is not None else None
            role = URIRef(r["r"]["value"])
            dm = r["dm"]["value"] if r.get("dm") is not None else None
            vi = r["vi"]["value"] if r.get("vi") is not None else None
            v = r["v"]["value"] if r.get("v") is not None else None
            cc = URIRef(r["cc"]["value"]) if r.get("cc") is not None else None
            atype = URIRef(r["atype"]["value"]) if r.get("atype") is not None else None

            artifacts_info[file] = {
                "main_entity": me,
                "role": role,
                "date_modified": dm,
                "version_iri": vi,
                "version_info": v,
                "file_size": None,
                "conformance_claim": cc,
                "additional_type": atype,
            }

    # get Version Indicators info only for Resources with certain Roles
    for k, v in artifacts_info.items():
        if v["role"] in [MRR.CatalogueData, MRR.ResourceData]:
            get_version_indicators_local(
                (manifest_path, manifest_root, manifest_graph), k, v
            )

    return artifacts_info


def artifact_file_name_from_graph_id(graph_id: str) -> str:
    s = graph_id.replace("://", "--")
    s = s.replace("/", "-")
    s = s.replace("#", "--")
    return s + ".ttl"


def store_remote_artifact_locally(
    manifest: Path | tuple[Path, Path, Graph],
    sparql_endpoint: str,
    graph_id: str,
    http_client: httpx.Client | None = None,
) -> Graph:
    """Writes a remote graph to a local file and registers that file as a Resource in the given Manifest.

    Only the Resource Role ResourceData is supported."""
    manifest_path, manifest_root, manifest_graph = get_manifest_paths_and_graph(
        manifest
    )
    q = """
        CONSTRUCT {
            ?s ?p ?o
        }
        WHERE {
            GRAPH <xxx> {
                ?s ?p ?o
            }
        }
        """.replace(
        "xxx", graph_id
    )
    r = query(sparql_endpoint, q, http_client)
    artifact_path = str(artifact_file_name_from_graph_id(graph_id))
    r.serialize(destination=manifest_root / artifact_path, format="longturtle")

    new_manifest_graph = Graph()
    new_manifest_graph += manifest_graph

    for m in new_manifest_graph.subjects(RDF.type, PREZ.Manifest):
        new_r = BNode()
        for r in new_manifest_graph.objects(m, PROF.hasResource):
            if (r, PROF.hasRole, MRR.ResourceData) in new_manifest_graph:
                new_r = r

        a = BNode()
        new_manifest_graph.add(
            (
                a,
                SDO.contentLocation,
                Literal(artifact_path),  # relative to manifest_root
            )
        )
        new_manifest_graph.add((a, SDO.mainEntity, URIRef(graph_id)))
        new_manifest_graph.add((new_r, PROF.hasArtifact, a))
        new_manifest_graph.add(
            (new_r, PROF.hasRole, MRR.ResourceData)  # only one supported for now
        )
        new_manifest_graph.add((m, PROF.hasResource, new_r))

    return new_manifest_graph


def update_local_artifact(
    manifest: Path | tuple[Path, Path, Graph],
    artifact_path: Path,
    sparql_endpoint: str,
    graph_id: str,
    http_client: httpx.Client | None = None,
):
    manifest_path, manifest_root, manifest_graph = get_manifest_paths_and_graph(
        manifest
    )
    q = """
        CONSTRUCT {
            ?s ?p ?o
        }
        WHERE {
            GRAPH <xxx> {
                ?s ?p ?o
            }
        }
        """.replace(
        "xxx", graph_id
    )
    r = query(sparql_endpoint, q, http_client)
    r.serialize(destination=artifact_path, format="longturtle")
